import fwdpy11
import numpy as np
import gvalue_recorder

N = 1000

pdict = {'demography': np.array([N]*100, dtype=np.uint32),
         'nregions': [],
         'sregions': [fwdpy11.GaussianS(0, 1, 1, 0.25)],
         'recregions': [fwdpy11.Region(0, 1, 1)],
         'rates': (0., 5e-3, 5e-3),
         'gvalue': fwdpy11.Additive(2., fwdpy11.GSS(VS=1, opt=1)),
         'prune_selected': False
         }

params = fwdpy11.ModelParams(**pdict)

pop = fwdpy11.DiploidPopulation(N, 1.0)

rng = fwdpy11.GSLrng(42)

gvalues = []
r = gvalue_recorder.record_gvalue(gvalues)
fwdpy11.evolvets(rng, pop, params, 100, r)

assert len(gvalues) == pop.generation, "Callback failure"



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import fwdpy11
import unittest
import pickle
import discrete_demography_roundtrips as ddr
import numpy as np


class TestMoveOrCopyIndividuals(unittest.TestCase):
    """
    Tests functions for object construction and
    object pickling
    """

    def test_move(self):
        m = fwdpy11.move_individuals(0, 2, 1, 0.1)
        self.assertEqual(m.when, 0)
        self.assertEqual(m.source, 2)
        self.assertEqual(m.destination, 1)
        self.assertEqual(m.number, 0)
        self.assertEqual(m.fraction, 0.1)
        self.assertEqual(m.move_individuals, True)
        self.assertEqual(m.resets_growth_rate, True)

        mp = pickle.dumps(m, -1)
        ump = pickle.loads(mp)
        self.assertEqual(ump.when, 0)
        self.assertEqual(ump.source, 2)
        self.assertEqual(ump.destination, 1)
        self.assertEqual(ump.number, 0)
        self.assertEqual(ump.fraction, 0.1)
        self.assertEqual(ump.move_individuals, True)
        self.assertEqual(ump.resets_growth_rate, True)
        self.assertTrue(m == ump)

    def test_copy(self):
        m = fwdpy11.copy_individuals(0, 2, 1, 0.1)
        self.assertEqual(m.when, 0)
        self.assertEqual(m.source, 2)
        self.assertEqual(m.destination, 1)
        self.assertEqual(m.number, 0)
        self.assertEqual(m.fraction, 0.1)
        self.assertEqual(m.move_individuals, False)

        mp = pickle.dumps(m, -1)
        ump = pickle.loads(mp)
        self.assertEqual(ump.when, 0)
        self.assertEqual(ump.source, 2)
        self.assertEqual(ump.destination, 1)
        self.assertEqual(ump.number, 0)
        self.assertEqual(ump.fraction, 0.1)
        self.assertEqual(ump.move_individuals, False)
        self.assertTrue(m == ump)


class TestSetDemeSize(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.ddsc = fwdpy11.SetDemeSize(0, 1, 3)

    def test_property_access(self):
        self.assertEqual(self.ddsc.when, 0)
        self.assertEqual(self.ddsc.deme, 1)
        self.assertEqual(self.ddsc.new_size, 3)
        self.assertEqual(self.ddsc.resets_growth_rate, True)

    def test_pickle(self):
        p = pickle.dumps(self.ddsc, -1)
        up = pickle.loads(p)
        self.assertEqual(self.ddsc.when, up.when)
        self.assertEqual(self.ddsc.deme, up.deme)
        self.assertEqual(self.ddsc.new_size, up.new_size)

    def test_invalid_deme(self):
        with self.assertRaises(ValueError):
            self.dpsc = fwdpy11.SetDemeSize(0, -1, 3)


class TestSetExponentialGrowth(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.seg = fwdpy11.SetExponentialGrowth(0, 1, 0.1)

    def test_property_access(self):
        self.assertEqual(self.seg.when, 0)
        self.assertEqual(self.seg.deme, 1)
        self.assertEqual(self.seg.G, 0.1)

    def test_pickle(self):
        p = pickle.dumps(self.seg, -1)
        up = pickle.loads(p)
        self.assertEqual(self.seg.when, up.when)
        self.assertEqual(self.seg.deme, up.deme)
        self.assertEqual(self.seg.G, up.G)

    def test_invalid_deme(self):
        with self.assertRaises(ValueError):
            self.seg = fwdpy11.SetExponentialGrowth(0, -1, 0.1)

    def test_invalid_growth_rate(self):
        with self.assertRaises(ValueError):
            self.seg = fwdpy11.SetExponentialGrowth(0, 1, np.nan)


class TestMigrationMatrix(unittest.TestCase):
    def testSimpleInit(self):
        fwdpy11.MigrationMatrix(np.identity(2))

    def testShape(self):
        m = fwdpy11.MigrationMatrix(np.identity(2))
        self.assertEqual(m.shape, (2, 2))

    def testPickle(self):
        m = fwdpy11.MigrationMatrix(np.identity(2))
        p = pickle.dumps(m, -1)
        up = pickle.loads(p)
        self.assertEqual(up.shape, (2, 2))
        self.assertTrue(np.array_equal(up.M, np.identity(2)))

    def testSettingRates(self):
        m = fwdpy11.MigrationMatrix(np.identity(2))
        m._set_migration_rates(0, [0.5, 0.5])
        m._set_migration_rates(1, np.array([0.01, 0.99]))
        self.assertTrue(np.array_equal(m.M, np.array(
            [0.5, 0.5, 0.01, 0.99]).reshape(2, 2)))

        m._set_migration_rates(np.array([0, 1, 2, 3]).reshape(2, 2))

        try:
            m._set_migration_rates(0, [2.5, 0.5])
        except:  # NOQA
            self.fail("unexpected exception")

        with self.assertRaises(ValueError):
            # Fails because deme index out of range
            m._set_migration_rates(2, [0.5, 0.5])

    def test_non_square_input(self):
        with self.assertRaises(ValueError):
            fwdpy11.MigrationMatrix(np.array([2.]*2))

        with self.assertRaises(ValueError):
            """ This matrix is also invalid b/c
            rows don't sum to 1, but we expect
            a failure before that check """
            fwdpy11.MigrationMatrix(
                np.array([i for i in range(6)]).reshape(2, 3))

    def test_weights_greater_than_one(self):
        with self.assertRaises(ValueError):
            fwdpy11.MigrationMatrix(np.array([2.]*4).reshape(2, 2), False)


class TestSetMigrationRates(unittest.TestCase):
    def test_init_from_list(self):
        m = fwdpy11.SetMigrationRates(0, 0, [0, 1, 0])
        self.assertEqual(m.when, 0)
        self.assertEqual(m.deme, 0)
        self.assertTrue(np.array_equal(m.migrates, np.array([0, 1, 0])))

    def test_init_from_numpy(self):
        m = fwdpy11.SetMigrationRates(0, 0, np.array([0, 1, 0]))
        self.assertEqual(m.when, 0)
        self.assertEqual(m.deme, 0)
        self.assertTrue(np.array_equal(m.migrates, np.array([0, 1, 0])))

    def test_reset_entire_matrix(self):
        m = fwdpy11.SetMigrationRates(3, np.identity(3))
        self.assertTrue(np.array_equal(m.migrates, np.identity(3)))

    def test_reset_entire_matrix_bad_inputs(self):
        with self.assertRaises(ValueError):
            fwdpy11.SetMigrationRates(3, np.arange(4).reshape(2, 2))

    def test_pickle(self):
        m = fwdpy11.SetMigrationRates(0, 0, np.array([0, 1, 0]))
        p = pickle.dumps(m, -1)
        up = pickle.loads(p)
        self.assertEqual(m.when, up.when)
        self.assertEqual(m.deme, up.deme)
        self.assertEqual(m.migrates.tolist(), up.migrates.tolist())


class TestDiscreteDemographyInitialization(unittest.TestCase):
    def test_init_migmatrix_with_tuple(self):
        mm = np.array([0.3, 0.7, 0.7, 0.3]).reshape(2, 2)
        try:
            d = fwdpy11.DiscreteDemography(migmatrix=mm)
            self.assertEqual(d.migmatrix.scaled, False)
        except:  # NOQA
            self.fail("unexpected exception")
        try:
            d = fwdpy11.DiscreteDemography(migmatrix=((mm, True)))
            self.assertEqual(d.migmatrix.scaled, True)
        except:  # NOQA
            self.fail("unexpected exception")
        try:
            d = fwdpy11.DiscreteDemography(
                migmatrix=((mm, False)))
            self.assertEqual(d.migmatrix.scaled, False)
        except:  # NOQA
            self.fail("unexpected exception")

    def test_move_or_copy(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        dd = fwdpy11.DiscreteDemography(m)
        self.assertTrue(dd.migmatrix is None)
        with self.assertRaises(AttributeError):
            del self.migmatrix

    def test_setting_growth(self):
        g = [fwdpy11.SetExponentialGrowth(0, 1, 0.3)]
        fwdpy11.DiscreteDemography(set_growth_rates=g)

    def test_setting_size_changes(self):
        c = [fwdpy11.SetDemeSize(0, 1, 15151)]
        fwdpy11.DiscreteDemography(set_deme_sizes=c)

    def test_init_from_numpy(self):
        """
        NOTE: this test is important b/c it tests
        that the __init__ function overloads can be seen.
        If they are re-ordered, then this test may fail.
        """
        N = 1000
        popsizes = np.array([N]*10*N + [int(0.1*N)] *
                            int(0.1*N) + [2*N]*100, dtype=np.uint32)
        demography = fwdpy11.DiscreteDemography(popsizes)
        self.assertEqual(len(np.unique(popsizes)),
                         len(demography.set_deme_sizes))

    def test_init_with_two_growth_changes_same_generation(self):
        g = [fwdpy11.SetExponentialGrowth(0, 1, 0.3)]*2
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(set_growth_rates=g)

    def test_init_with_two_selfing_changes_same_generation(self):
        g = [fwdpy11.SetSelfingRate(0, 1, 0.3)]*2
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(set_selfing_rates=g)

    def test_init_with_two_deme_size_changes_same_generation(self):
        c = [fwdpy11.SetDemeSize(0, 1, 15151)]*2
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(set_deme_sizes=c)

    def test_init_with_two_mass_moves_same_generation(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 0.5)]*2
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(m)

    def test_init_with_two_mass_copies_same_generation(self):
        m = [fwdpy11.copy_individuals(0, 0, 1, 0.5)]*2
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(m)

    def test_init_with_one_mass_move_and_one_mass_copy_same_generation(self):
        # Internally, the input data will get sorted
        # so that copies happen before moves.
        m = [fwdpy11.move_individuals(0, 0, 1, 0.5),
             fwdpy11.copy_individuals(0, 0, 1, 0.5)]
        try:
            d = fwdpy11.DiscreteDemography(m)
            self.assertEqual(
                d.mass_migrations[0].move_individuals, False)
            self.assertEqual(d.mass_migrations[1].move_individuals, True)
        except:  # NOQA
            self.fail("unexpected exception")

    def test_move_too_many_individuals_from_same_deme(self):
        # Attempt to move 125% of deme 0
        m = [fwdpy11.move_individuals(0, 0, 1, 0.5),
             fwdpy11.move_individuals(0, 0, 2, 0.75)]
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(m)

    def test_move_too_many_individuals_from_same_deme_with_copy_in_middle(self):
        # Attempt to move 125% of deme 0
        # This is a test that input data get sorted so
        # that copies are before moves, which will allow
        # detecting the invalid cumulative move attempt
        m = [fwdpy11.move_individuals(0, 0, 1, 0.5),
             fwdpy11.copy_individuals(0, 0, 1, 0.5),
             fwdpy11.move_individuals(0, 0, 2, 0.75)]
        with self.assertRaises(ValueError):
            fwdpy11.DiscreteDemography(m)

    def test_with_migration_matrix(self):
        mm = np.array([0.72, 0.28, 0.3, 0.7]).reshape(2, 2)
        d = fwdpy11.DiscreteDemography(migmatrix=mm)
        p = pickle.dumps(d, -1)
        up = pickle.loads(p)
        self.assertTrue(np.array_equal(up.migmatrix.M, mm))
        with self.assertRaises(AttributeError):
            del d.migmatrix

    def test_identity_matrix_conversion_to_None(self):
        mm = np.identity(5)
        d = fwdpy11.DiscreteDemography(migmatrix=mm)
        self.assertTrue(d.migmatrix is None)

    def test_identity_matrix_with_migrate_changes(self):
        mm = np.identity(5)
        cm = [fwdpy11.SetMigrationRates(0, 3, [0.2]*5)]
        d = fwdpy11.DiscreteDemography(migmatrix=mm,
                                       set_migration_rates=cm)
        self.assertTrue(d.migmatrix is not None)


class TestDiscreteDemography(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.rng = fwdpy11.GSLrng(42)
        self.pop = fwdpy11.DiploidPopulation(100)

    def test_simple_moves_from_single_deme(self):
        d = fwdpy11.DiscreteDemography(
            [fwdpy11.move_individuals(0, 0, 1, 0.5)])
        ddr.DiscreteDemography_roundtrip(self.rng,
                                         self.pop, d, 1)
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(len(deme_sizes[0]), 2)
        for i in range(len(deme_sizes[0])):
            self.assertEqual(deme_sizes[1][i], 50)

    def test_simple_copies_from_single_deme(self):
        d = fwdpy11.DiscreteDemography(
            [fwdpy11.copy_individuals(0, 0, 1, 0.5)])
        ddr.DiscreteDemography_roundtrip(self.rng,
                                         self.pop, d, 1)
        deme_sizes = self.pop.deme_sizes()
        expected = {0: 100, 1: 50}
        self.assertEqual(len(deme_sizes[0]), len(expected))
        for i in range(len(deme_sizes[0])):
            self.assertEqual(deme_sizes[1][i], expected[i])

    def test_simple_back_and_forth_move(self):
        d = fwdpy11.DiscreteDemography(
            [fwdpy11.move_individuals(1, 0, 1, 0.5),
             fwdpy11.move_individuals(2, 1, 0, 1.)])

        ddr.DiscreteDemography_roundtrip(self.rng,
                                         self.pop, d, 3)
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(len(deme_sizes[0]), 1)
        for i in range(len(deme_sizes[0])):
            self.assertEqual(deme_sizes[1][i], self.pop.N)

    def test_simple_back_and_forth_copy(self):
        d = fwdpy11.DiscreteDemography(
            [fwdpy11.copy_individuals(1, 0, 1, 0.5),
             fwdpy11.copy_individuals(2, 1, 0, 1.)])

        ddr.DiscreteDemography_roundtrip(self.rng,
                                         self.pop, d, 3)
        expected = {0: 150, 1: 50}
        self.assertEqual(expected,
                         self.pop.deme_sizes(as_dict=True))

    def test_simple_moves_from_multiple_demes(self):
        # In generation 0, create deme 1 from 0.
        # In generation 1, we move 1/2 of deme 1 to
        # deme 2, which creates a new deme:
        d = fwdpy11.DiscreteDemography(
            [fwdpy11.move_individuals(0, 0, 1, 0.5),
             fwdpy11.move_individuals(1, 1, 2, 0.5)])

        # Evolve for two generations:
        ddr.DiscreteDemography_roundtrip(self.rng,
                                         self.pop, d, 2)
        # We now expect 50, 25, and 25 individuals in
        # demes 0, 1, and 2
        expected = {0: 50, 1: 25, 2: 25}
        self.assertEqual(expected, self.pop.deme_sizes(as_dict=True))

    def test_simple_copies_from_multiple_demes(self):
        # Set 1/2 the population to start in deme 1.
        # In generation 0, we move 1/2 of deme 1 to
        # deme 2, which creates a new deme:
        d = fwdpy11.DiscreteDemography(
            [fwdpy11.move_individuals(0, 0, 1, 0.5),
             fwdpy11.copy_individuals(1, 1, 2, 0.5)])
        # Evolve for one generation:
        ddr.DiscreteDemography_roundtrip(self.rng,
                                         self.pop, d, 2)
        # We now expect 50, 50, and 25 individuals in
        # demes 0, 1, and 2
        expected = {0: 50, 1: 50, 2: 25}
        self.assertEqual(expected, self.pop.deme_sizes(as_dict=True))

    def test_single_deme_growth(self):
        N1 = 3412
        t = 111
        G = np.exp((np.log(N1)-np.log(self.pop.N))/t)
        g = [fwdpy11.SetExponentialGrowth(16, 0, G)]
        d = fwdpy11.DiscreteDemography(set_growth_rates=g)
        ddr.DiscreteDemography_roundtrip(self.rng,
                                         self.pop, d, 15 + t + 1)
        self.assertEqual(self.pop.N, N1)
        self.assertEqual(len(self.pop.diploid_metadata), N1)

    def test_two_deme_growth(self):
        N0 = [90, 10]
        t = [14, 23]  # generations of growth in each deme
        N1 = [5361, 616]
        G0 = np.exp((np.log(N1[0])-np.log(N0[0]))/t[0])
        G1 = np.exp((np.log(N1[1])-np.log(N0[1]))/t[1])

        g = []
        g.append(fwdpy11.SetExponentialGrowth(7, 0, G0))
        g.append(fwdpy11.SetExponentialGrowth(7+t[0], 0, fwdpy11.NOGROWTH))
        g.append(fwdpy11.SetExponentialGrowth(33, 1, G1))
        g.append(fwdpy11.SetExponentialGrowth(33+t[1], 1, fwdpy11.NOGROWTH))
        moves = [fwdpy11.move_individuals(0, 0, 1, 0.1)]
        d = fwdpy11.DiscreteDemography(set_growth_rates=g,
                                       mass_migrations=moves)
        ddr.DiscreteDemography_roundtrip(self.rng, self.pop,
                                         d, 100)

        self.assertEqual(self.pop.N, sum(N1))
        deme_sizes = self.pop.deme_sizes()
        for i, j in zip(deme_sizes[1], N1):
            self.assertEqual(i, j)

    def test_two_deme_growth_with_hard_reset(self):
        N0 = [90, 10]
        t = [14, 23]  # generations of growth in each deme
        N1 = [5361, 616]
        G0 = np.exp((np.log(N1[0])-np.log(N0[0]))/t[0])
        G1 = np.exp((np.log(N1[1])-np.log(N0[1]))/t[1])

        g = []
        g.append(fwdpy11.SetExponentialGrowth(7, 0, G0))
        g.append(fwdpy11.SetExponentialGrowth(33, 1, G1))
        g.append(fwdpy11.SetExponentialGrowth(33+t[1], 1, fwdpy11.NOGROWTH))
        # Cut off the growth in deme 0 after a few generations,
        # and manually set the new deme size to 100 w/no growth
        p = [fwdpy11.SetDemeSize(11, 0, 100)]
        moves = [fwdpy11.move_individuals(0, 0, 1, 0.1)]
        d = fwdpy11.DiscreteDemography(set_deme_sizes=p,
                                       set_growth_rates=g,
                                       mass_migrations=moves)
        ddr.DiscreteDemography_roundtrip(self.rng, self.pop,
                                         d, 100)

        deme_sizes = self.pop.deme_sizes()
        N1 = [100, N1[1]]
        self.assertEqual(self.pop.N, sum(N1))
        for i, j in zip(deme_sizes[1], N1):
            self.assertEqual(i, j)

    def test_two_deme_growth_without_hard_reset(self):
        N0 = [90, 10]
        t = [14, 23]  # generations of growth in each deme
        N1 = [5361, 616]
        G0 = np.exp((np.log(N1[0])-np.log(N0[0]))/t[0])
        G1 = np.exp((np.log(N1[1])-np.log(N0[1]))/t[1])

        g = []
        g.append(fwdpy11.SetExponentialGrowth(7, 0, G0))
        g.append(fwdpy11.SetExponentialGrowth(7+t[0], 0, fwdpy11.NOGROWTH))
        g.append(fwdpy11.SetExponentialGrowth(33, 1, G1))
        g.append(fwdpy11.SetExponentialGrowth(33+t[1], 1, fwdpy11.NOGROWTH))
        # after X generations of growth, N[0] changes to 100
        # and the growth rate is not reset.
        p = [fwdpy11.SetDemeSize(11, 0, 100, False)]
        moves = [fwdpy11.move_individuals(0, 0, 1, 0.1)]
        d = fwdpy11.DiscreteDemography(set_growth_rates=g,
                                       set_deme_sizes=p,
                                       mass_migrations=moves)
        ddr.DiscreteDemography_roundtrip(self.rng, self.pop,
                                         d, 100)

        deme_sizes = self.pop.deme_sizes()
        N1[0] = np.round(100.*np.power(G0, 7 + t[0] - 11))
        self.assertEqual(self.pop.N, sum(N1))
        for i, j in zip(deme_sizes[1], N1):
            self.assertEqual(i, j)

    def test_two_moves_in_same_generation(self):
        """
        In generation 0, deme 0 splits equally
        into demes 1 and 2.  This should leave
        deme 0 empty and the sizes of demes
        1 and 2 both equal to 0.5 the initial
        size.
        """
        m = [fwdpy11.move_individuals(0, 0, 2, 0.5),
             fwdpy11.move_individuals(0, 0, 1, 0.5)]
        d = fwdpy11.DiscreteDemography(m)
        N0 = self.pop.N
        ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 1)
        expected = {1: N0//2, 2: N0//2}
        self.assertEqual(expected, self.pop.deme_sizes(as_dict=True))

    def test_copies_happen_before_moves(self):
        """
        This is a more complex test.
        In the same generation:
        1. All individuals are copied from deme 0 to deme 1.
        2. 50% of deme 0 moves to deme 2.
        3. 50% of deme 0 moves to deme 3.

        This, at the end, the size of deme 1 should be
        the initial size and the size of demes 2 and 3
        should be 0.5*initial size
        """
        m = [fwdpy11.move_individuals(0, 0, 3, 0.5),
             fwdpy11.copy_individuals(0, 0, 1, 1),
             fwdpy11.move_individuals(0, 0, 2, 0.5)]
        d = fwdpy11.DiscreteDemography(m)
        N0 = self.pop.N
        ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 1)
        expected = {1: N0, 2: N0//2, 3: N0//2}
        self.assertEqual(expected, self.pop.deme_sizes(as_dict=True))

    def test_mass_move_with_growth(self):
        """
        In generation 5, the mass movement from 0 to 1
        will reset the growth rate in deme 0 to
        fwdpy11.NOGROWTH

        This test is also handy b/c growth results in
        an odd total N by the time the mass migration happens
        """
        m = [fwdpy11.move_individuals(5, 0, 1, 0.5)]
        g = [fwdpy11.SetExponentialGrowth(0, 0, 1.1)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m, set_growth_rates=g)
        N0 = self.pop.N
        ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 10)
        deme_sizes = self.pop.deme_sizes()
        N5 = np.round(N0*np.power(g[0].G, 5))
        N_after_mass_mig = [N5//2, N5-N5//2]
        for i, j in zip(deme_sizes[1], N_after_mass_mig):
            self.assertEqual(i, j)

    def test_mass_move_with_growth_no_reset(self):
        """
        In generation 5, the mass movement from 0 to 1
        """
        m = [fwdpy11.move_individuals(5, 0, 1, 0.5, False)]
        g = [fwdpy11.SetExponentialGrowth(0, 0, 1.1)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m, set_growth_rates=g)
        N0 = self.pop.N
        ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 10)
        deme_sizes = self.pop.deme_sizes()
        N5 = np.round(N0*np.power(g[0].G, 5))
        N_after_mass_mig_0 = np.round((N5//2)*np.power(g[0].G, 5))
        N = [N_after_mass_mig_0, N5 - N5//2]
        for i, j in zip(N, deme_sizes[1]):
            self.assertEqual(i, j)

    def test_migration_matrix_too_small(self):
        """
        The mig matrix is 2x2 but a mass migration event
        tries to create a 3rd deme, which triggers an exception
        """
        mm = np.array([0.5]*4).reshape(2, 2)
        m = [fwdpy11.move_individuals(0, 0, 1, 0.5),
             fwdpy11.move_individuals(0, 0, 2, 0.5)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m, migmatrix=mm)
        with self.assertRaises(ValueError):
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)

    def test_simple_two_deme_migration(self):
        mm = np.array([0.5]*4).reshape(2, 2)
        m = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        d = fwdpy11.DiscreteDemography(migmatrix=mm, mass_migrations=m)
        N = self.pop.N
        ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)
        self.assertEqual(N, self.pop.N)
        deme_sizes = self.pop.deme_sizes()
        for i in deme_sizes[1]:
            self.assertEqual(i, self.pop.N//2)

    def test_change_migration_rates_simple_two_deme_migration(self):
        """
        For a 2-deme model, the mig matrix is
        [[0, 1]
         [1, 0]]
        so that all offspring have both parents from the other deme,
        which gives us an easy check on how many migration events
        will be recorded by the test simulation.

        After 3 generations, we reset the migration rates to be
        [[1, 0],
         [1, 0]]
        so that all parents are from deme zero.
        """
        mm = np.array([0, 1, 1, 0]).reshape(2, 2)
        mmigs = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        smr = [fwdpy11.SetMigrationRates(
            3, np.array([1, 0, 1, 0]).reshape(2, 2))]
        d = fwdpy11.DiscreteDemography(mass_migrations=mmigs, migmatrix=mm,
                                       set_migration_rates=smr)
        N = self.pop.N
        migevents = ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)
        self.assertEqual(N, self.pop.N)
        self.assertEqual(len(migevents), 2*N*3 + 2*N)
        deme_sizes = self.pop.deme_sizes()
        for i in deme_sizes[1]:
            self.assertEqual(i, self.pop.N//2)

    def test_change_migration_rates_simple_two_deme_migration_bad_matrix(self):
        """
        For a 2-deme model, the mig matrix is
        [0, 1
         1, 0]
        so that all offspring have both parents from the other deme,
        which gives us an easy check on how many migration events
        will be recorded by the test simulation.

        After 3 generations, we reset the migration rates to be
        [[0.5, 0.5],
         [0, 0]],
        which leads to there being no parents for deme 1, raising a
        fwdpy11.DemographyError exception.
        """
        mm = np.array([0, 1, 1, 0]).reshape(2, 2)
        mmigs = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        smr = [fwdpy11.SetMigrationRates(
            3, np.array([0.5, 0.5, 0, 0]).reshape(2, 2))]
        d = fwdpy11.DiscreteDemography(mass_migrations=mmigs, migmatrix=mm,
                                       set_migration_rates=smr)
        with self.assertRaises(fwdpy11.DemographyError):
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)

    def test_selfing_vs_migration(self):
        """
        Parents of deme 0 are all migrants from deme 1.
        Parents of deme 1 are all migrants from deme 0.
        Deme 0 never selfs.  Deme 1 always selfs.
        """
        mm = np.array([0, 1, 1, 0]).reshape(2, 2)
        mmigs = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        s = [fwdpy11.SetSelfingRate(0, 1, 1.0)]
        mmigs = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        d = fwdpy11.DiscreteDemography(
            mass_migrations=mmigs, migmatrix=mm, set_selfing_rates=s)
        migevents = ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)
        for m in migevents:
            if m[1] == 1:  # parent is from deme 1
                self.assertTrue(m[3] == ddr.MatingEventType.selfing)
                self.assertNotEqual(m[1], m[2])  # Must be a migrant parent
            elif m[1] == 0:
                self.assertTrue(m[3] == ddr.MatingEventType.outcrossing)
                self.assertNotEqual(m[1], m[2])  # Must be a migrant parent
            else:
                self.fail("invalid parental deme type")


class TestMigrationModels(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.rng = fwdpy11.GSLrng(12352531)

    def test_migration_matrix_too_large(self):
        """
        Two demes and a 3x3 matrix
        """
        pop = fwdpy11.DiploidPopulation([20, 20], 1.)
        mm = np.identity(3)
        mm[:] = 0.25
        np.fill_diagonal(mm, 0)
        np.fill_diagonal(mm, 1-np.sum(mm, 1))
        M = fwdpy11.MigrationMatrix(mm)
        d = fwdpy11.DiscreteDemography(migmatrix=M)
        with self.assertRaises(ValueError):
            ddr.DiscreteDemography_roundtrip(self.rng, pop, d, 5)


class TestGhostPopulations(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(100, 1.)
        self.rng = fwdpy11.GSLrng(4321987)

    def test_simple_ghost_pop_example(self):
        mmigs = [fwdpy11.copy_individuals(5, 0, 1, 1.0),
                 fwdpy11.copy_individuals(10, 1, 0, 0.10)]
        size_changes = [fwdpy11.SetDemeSize(10, 1, 0),
                        fwdpy11.SetDemeSize(10, 0, 100)]
        dd = fwdpy11.DiscreteDemography(mass_migrations=mmigs,
                                        set_deme_sizes=size_changes)
        ddr.DiscreteDemography_roundtrip(self.rng, self.pop, dd, 20)
        self.assertEqual(self.pop.N, 100)
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(len(deme_sizes[0]), 1)
        self.assertEqual(deme_sizes[0][0], 0)
        self.assertEqual(deme_sizes[1][0], 100)


class TestExponentialDecline(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(100, 1.)
        self.rng = fwdpy11.GSLrng(3321986)

    def test_global_extinction_single_deme(self):
        g = [fwdpy11.SetExponentialGrowth(0, 0, 0.5)]
        dd = fwdpy11.DiscreteDemography(set_growth_rates=g)
        with self.assertRaises(fwdpy11.GlobalExtinction):
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, dd, 20)


class TestDemographyError(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(100, 1.)
        self.rng = fwdpy11.GSLrng(3321986)

    def test_growth_in_deme_that_doesnt_exist(self):
        g = [fwdpy11.SetExponentialGrowth(0, 1, 0.5)]
        dd = fwdpy11.DiscreteDemography(set_growth_rates=g)
        with self.assertRaises(fwdpy11.DemographyError):
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, dd, 20)

    def test_growth_in_deme_that_went_extinct(self):
        mm = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        d = [fwdpy11.SetDemeSize(5, 1, 0)]
        g = [fwdpy11.SetExponentialGrowth(6, 1, 0.5)]
        dd = fwdpy11.DiscreteDemography(set_growth_rates=g,
                                        mass_migrations=mm,
                                        set_deme_sizes=d)
        with self.assertRaises(fwdpy11.DemographyError):
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, dd, 20)

    def test_set_selfing_in_deme_that_doesnt_exist(self):
        g = [fwdpy11.SetSelfingRate(0, 1, 0.5)]
        dd = fwdpy11.DiscreteDemography(set_selfing_rates=g)
        with self.assertRaises(fwdpy11.DemographyError):
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, dd, 20)

    def test_set_selfing_in_deme_that_went_extinct(self):
        mm = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        d = [fwdpy11.SetDemeSize(5, 1, 0)]
        g = [fwdpy11.SetSelfingRate(6, 1, 0.5)]
        dd = fwdpy11.DiscreteDemography(set_selfing_rates=g,
                                        mass_migrations=mm,
                                        set_deme_sizes=d)
        with self.assertRaises(fwdpy11.DemographyError):
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, dd, 20)

    def test_migration_into_empty_deme(self):
        mass_mig = [fwdpy11.move_individuals(0, 0, 2, 0.5)]
        mm = np.zeros(9).reshape(3, 3)
        mm[:, 2] = 1.
        np.fill_diagonal(mm, 0)
        np.fill_diagonal(mm, 1-np.sum(mm, 1))
        M = fwdpy11.MigrationMatrix(mm)
        d = fwdpy11.DiscreteDemography(migmatrix=M, mass_migrations=mass_mig)
        assert_raised = False
        try:
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)
        except fwdpy11.DemographyError as e:
            self.assertTrue("rate into empty destination deme" in str(e))
            self.assertTrue("from empty parental deme" not in str(e))
            assert_raised = True
        self.assertTrue(assert_raised)

    def test_migration_from_an_empty_deme(self):
        mass_mig = [fwdpy11.move_individuals(0, 0, 2, 0.5)]
        mm = np.identity(3)
        mm[:] = 0.25
        np.fill_diagonal(mm, 0)
        np.fill_diagonal(mm, 1-np.sum(mm, 1))
        M = fwdpy11.MigrationMatrix(mm)
        d = fwdpy11.DiscreteDemography(migmatrix=M, mass_migrations=mass_mig)
        assert_raised = False
        try:
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)
        except fwdpy11.DemographyError as e:
            self.assertTrue("from empty parental deme" in str(e))
            self.assertTrue("into empty destination" not in str(e))
            assert_raised = True
        self.assertTrue(assert_raised)

    def test_migration_from_an_empty_deme_into_an_empty_deme(self):
        mass_mig = [fwdpy11.move_individuals(0, 0, 2, 1.0)]
        mm = np.array([0., 1., 0., 0., 0., 0., 0., 0., 1.]).reshape(3, 3)
        d = fwdpy11.DiscreteDemography(migmatrix=mm, mass_migrations=mass_mig)
        assert_raised = False
        try:
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)
        except fwdpy11.DemographyError as e:
            self.assertTrue("from empty parental deme" in str(e))
            self.assertTrue("into empty destination" in str(e))
            assert_raised = True
        self.assertTrue(assert_raised)

    def test_no_valid_parents(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s)
        assert_raised = False
        try:
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)
        except fwdpy11.DemographyError:
            assert_raised = True
        self.assertTrue(assert_raised)

    def test_no_valid_parents_alternate_method(self):
        m = [fwdpy11.copy_individuals(0, 0, 1, 1.0)]
        s = [fwdpy11.SetDemeSize(0, 0, 0), fwdpy11.SetDemeSize(2, 0, 100)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m, set_deme_sizes=s)
        assert_raised = False
        try:
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)
        except fwdpy11.DemographyError:
            assert_raised = True
        self.assertTrue(assert_raised)

    def test_no_valid_parents_with_migration(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5]*4).reshape(2, 2)
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s,
                                       migmatrix=M)
        assert_raised = False
        try:
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)
        except fwdpy11.DemographyError:
            assert_raised = True
        self.assertTrue(assert_raised)

    def test_no_valid_parents_with_migration_2(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5]*4).reshape(2, 2)
        cM = [fwdpy11.SetMigrationRates(
            0, np.array([0, 1, 0, 1]).reshape(2, 2))]
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s,
                                       set_migration_rates=cM,
                                       migmatrix=M)
        try:
            ddr.DiscreteDemography_roundtrip(self.rng, self.pop, d, 5)
        except Exception:
            self.fail("unexpected exception")


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import fwdpy11
import fwdpy11.demographic_models.IM
import argparse
import sys
import numpy as np
import concurrent.futures
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
HAVE_MOMENTS = False
try:
    import moments
    HAVE_MOMENTS = True
except ImportError:
    import warnings
    warnings.warn("Couldn't import moments, so we won't be making any plots!")
    pass


def make_parser():
    ADHF = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser("IM.py", formatter_class=ADHF)
    parser.add_argument('--split', '-s', type=float,
                        help="Fraction that splits into deme 0")
    parser.add_argument('--tsplit', '-T', type=float,
                        help="Time since population split,"
                        " in units of 2*Nref generations")
    parser.add_argument('--seed', type=int, help="Random number seed")

    optional = parser.add_argument_group("Optional")
    optional.add_argument('--Nref', type=int, default=1000,
                          help="Ancestral population size")
    optional.add_argument('--N0', type=float, default=1.0,
                          help="Contemporary size of deme 0, relative to Nref")
    optional.add_argument('--N1', type=float, default=1.0,
                          help="Contemporary size of deme 1, relative to Nref")
    optional.add_argument('--gamma', type=float, help="2Ns")
    # Require 0 <= h <= 2
    optional.add_argument('-H', type=float, default=1.0, help="Dominance")
    optional.add_argument('--nreps', type=int, default=1,
                          help="Number of forward simulation replicates")
    optional.add_argument('--migrates', '-M', type=float,
                          nargs=2, default=[0., 0.],
                          help="Migration rates, scaled by 2*Nref.")
    optional.add_argument('--nsam', type=int, default=15,
                          help="Number of diploids to sample from each deme")
    optional.add_argument('--theta', type=float, default=1.0,
                          help="Scaled mutation rate. "
                          "Danger zone with selection :).")
    optional.add_argument('--rho', type=float, default=1e3,
                          help="Scaled recombination rate, rho = 4*Nref*r")
    optional.add_argument('--show2d', action='store_true', default=False,
                          help="Display 2d fs comparison of the mean fs from"
                               "the simulation to moments")

    return parser


def IM_moments(params, ns, gamma=0.0, h=0.5):
    """
    Expected FS for IM model with selection

    ns: sample sizes, given as [ns1, ns2]
    s: frac that splits into deme 1 (1-s into deme 2)
    nu1/nu2: contemporary population sizes, with exponenential size change
    T: time between split and now (in 2Nref generations)
    m12/m21: migration rates, scaled by 2Nref
             mij is rate from j into i

    The mutation rate theta=4*N_ref*u is assumed to be 1.

    ns: the sample sizes (don't have to be equal) given as list of length 2

    gamma: pop-size scaled selection coefficient (2Ns), default 0
    h: dominance coefficient, default 0.5

    Note that when integrating, gamma and h must be passed as lists of same
    length as number of demes, with each entry specifying the coefficient in
    each deme. If they are all the same, pass
        `gamma=[gamma, gamma, ..., gamma], h=[h, h, ..., h]`
    where the lengths are equal to the number of demes.

    gamma = 2*N_ref*s, with the interpretation of fitnesses:
        aa : 1
        Aa : 1+2hs
        AA : 1+2s
    """
    s, nu1, nu2, T, m12, m21 = params
    # equilibrium frequency spectrum
    sts = moments.LinearSystem_1D.steady_state_1D(
        ns[0] + ns[1], gamma=gamma, h=h)
    fs = moments.Spectrum(sts)
    # split into two demes
    fs = moments.Manips.split_1D_to_2D(fs, ns[0], ns[1])
    # define size change function
    A, B = 1., 1.
    if s is not None:
        A, B = 1.0 - s, s

    def nu1_func(t): return A * np.exp(np.log(nu1/A)*t/T)

    def nu2_func(t):
        return B * np.exp(np.log(nu2/B)*t/T)

    def nu_func(t): return [nu1_func(t), nu2_func(t)]
    # integrate for time T
    fs.integrate(nu_func, T, m=np.array([[0, m12], [m21, 0]]),
                 gamma=[gamma, gamma], h=[h, h])
    return fs


def build_demography(args):
    """
    Returns the demography, the simulation length, and the
    final total size in each deme.

    The main real work here is to convert the input
    paremeters to match the scaling expected by
    fwdpy11.demographic_models.IM.two_deme_IM:

    1. The input migration rates are in units of 2*Nref
    2. The time since the split is also in units of 2*Nref
    """
    two_deme_IM = fwdpy11.demographic_models.IM.two_deme_IM

    # Rescale input migration rates
    migrates = tuple(i/(2.*args.Nref) for i in args.migrates)

    # Change split time from generations/(2*Nref) to
    # generations/Nref.
    d, t1, t2 = two_deme_IM(args.Nref, 2.0*args.tsplit, args.split,
                            (args.N0, args.N1), migrates,
                            burnin=20.)
    simlen = int(t1 + t2)
    N0 = np.rint(args.N0*args.Nref).astype(int)
    N1 = np.rint(args.N1*args.Nref).astype(int)
    return d, simlen, (N0, N1)


def build_parameters_dict(args):
    """
    Returns sim params and the final sizes
    in each deme
    """
    demog, simlen, finalNs = build_demography(args)

    nregions = []
    sregions = []
    recregions = [fwdpy11.PoissonInterval(0, 1, args.rho/(4.*args.Nref))]

    rates = (0, 0, None)
    if args.gamma is not None:
        sregions = [fwdpy11.ConstantS(
            0, 1, 1, args.gamma, args.H, scaling=2*args.Nref)]
        rates = (0, args.theta/(4.0*args.Nref), None)

    pdict = {'nregions': nregions,
             'sregions': sregions,
             'recregions': recregions,
             'rates': rates,
             'gvalue': fwdpy11.Multiplicative(2.),
             'demography': demog,
             'simlen': simlen,
             'prune_selected': True
             }

    return pdict, finalNs


def full_joint_fs(pop):
    """
    Get the marginal SFS per deme.
    """
    nt = np.array(pop.tables.nodes, copy=False)
    for i in pop.diploid_metadata:
        assert i.deme == nt['deme'][i.nodes[0]]
        assert i.deme == nt['deme'][i.nodes[1]]
    an = pop.alive_nodes
    deme0 = an[np.where(nt['deme'][an] == 0)[0]]
    deme1 = an[np.where(nt['deme'][an] == 1)[0]]

    jfs = pop.tables.fs([deme0, deme1], marginalize=False)
    return jfs


def subsample_fs(pop, args):
    md = np.array(pop.diploid_metadata, copy=False)
    nodes0 = md['nodes'][np.where(md['deme'] == 0)[0]].flatten()
    nodes1 = md['nodes'][np.where(md['deme'] == 1)[0]].flatten()

    s0 = np.random.choice(nodes0, 2*args.nsam, replace=False)
    s1 = np.random.choice(nodes1, 2*args.nsam, replace=False)

    fs = pop.tables.fs([s0, s1], marginalize=True)
    return fs[0][1:-1], fs[1][1:-1]


def project_fs(sfs, n):
    fs = moments.Spectrum(sfs)
    psfs = fs.project([n])
    return psfs.data[1:-1]


def runsim(args, pdict, seed):
    pop = fwdpy11.DiploidPopulation(args.Nref, 1.0)
    rng = fwdpy11.GSLrng(seed)
    np.random.seed(seed)

    params = fwdpy11.ModelParams(**pdict)
    fwdpy11.evolvets(rng, pop, params, 100)
    if args.gamma is None:
        fwdpy11.infinite_sites(rng, pop, args.theta/float(4.*args.Nref))
    full_jfs = full_joint_fs(pop)
    sfs0_rs, sfs1_rs = subsample_fs(pop, args)
    return full_jfs, sfs0_rs, sfs1_rs


def plot_fs(args, moments_fs, fwdpy11_fs,
            fwdpy11_sample_fs):
    fwdpy11_fs0 = project_fs(fwdpy11_fs[0], 2*args.nsam)
    fwdpy11_fs1 = project_fs(fwdpy11_fs[1], 2*args.nsam)
    moments_fs0 = moments_fs.marginalize([1])
    moments_fs1 = moments_fs.marginalize([0])

    fig = plt.figure(figsize=(7, 5))
    gs = gridspec.GridSpec(ncols=2, nrows=1, figure=fig)
    deme0 = fig.add_subplot(gs[0, 0])
    deme1 = fig.add_subplot(gs[0, 1])
    x = [i for i in range(2*args.nsam-1)]
    deme0.plot(x, fwdpy11_fs0, 'bo',
               label="fwdpy11 projected", alpha=0.2, zorder=2)
    deme0.plot(x, fwdpy11_sample_fs[0], 'go',
               label="fwdpy11 sample", alpha=0.2, zorder=3)
    deme0.plot(x, args.theta*moments_fs0.data[1:-1],
               'r-', alpha=0.2, label="moments", zorder=1)
    deme0.plot(x, args.theta*moments_fs0.data[1:-1], 'r+', zorder=1)

    deme1.plot(x, fwdpy11_fs1, 'bo',
               label="fwdpy11 projected", alpha=0.2, zorder=2)
    deme1.plot(x, fwdpy11_sample_fs[1], 'go',
               label="fwdpy11 sample", alpha=0.2,  zorder=3)
    deme1.plot(x, args.theta*moments_fs1.data[1:-1],
               'r-', alpha=0.2, label="moments", zorder=1)
    deme1.plot(x, args.theta*moments_fs1.data[1:-1], 'r+', zorder=1)
    deme0.set_title("Deme 0")
    deme1.set_title("Deme 1")
    deme0.legend()
    deme1.legend()
    deme0.set_xlabel("Derived frequency")
    deme1.set_xlabel("Derived frequency")
    deme0.set_ylabel("E[# mutations]")
    plt.tight_layout()
    plt.savefig("moments.png")


if __name__ == "__main__":
    parser = make_parser()
    args = parser.parse_args(sys.argv[1:])

    if HAVE_MOMENTS is True:
        moments_params = (args.split, args.N0, args.N1, args.tsplit,
                          args.migrates[1], args.migrates[0])
        moments_nsam = (2*args.nsam, 2*args.nsam)
        mgamma = args.gamma
        if mgamma is None:
            mgamma = 0.0
        moments_fs = IM_moments(moments_params, moments_nsam,
                                mgamma, args.H/2.0)

    # Fail early if input params are bad
    pdict, finalNs = build_parameters_dict(args)
    np.random.seed(args.seed)
    seeds = np.random.randint(0, np.iinfo(np.uint32).max, args.nreps)

    sum_fs0 = np.zeros(2*finalNs[0]+1)
    sum_fs1 = np.zeros(2*finalNs[1]+1)
    sum_samples_fs0 = np.zeros(2*args.nsam-1)
    sum_samples_fs1 = np.zeros(2*args.nsam-1)
    mean_full_jFS = None
    with concurrent.futures.ProcessPoolExecutor() as e:
        futures = {e.submit(runsim, args, pdict, i) for i in seeds}
        for fut in concurrent.futures.as_completed(futures):
            full_jfs, sfs0_rs, sfs1_rs = fut.result()
            if mean_full_jFS is None:
                mean_full_jFS = full_jfs
            else:
                mean_full_jFS += full_jfs
            sum_fs0 += full_jfs.sum(axis=1).todense()
            sum_fs1 += full_jfs.sum(axis=0).todense()
            sum_samples_fs0 += sfs0_rs
            sum_samples_fs1 += sfs1_rs

    # NOTE: using /= for the next
    # line basically zeros everything out!
    mean_full_jFS = mean_full_jFS/args.nreps
    sum_fs0 /= args.nreps
    sum_fs1 /= args.nreps
    sum_samples_fs0 /= args.nreps
    sum_samples_fs1 /= args.nreps
    if HAVE_MOMENTS is True:
        mean_full_jFS = moments.Spectrum(mean_full_jFS.todense())
        projected_jfs = mean_full_jFS.project([2*args.nsam, 2*args.nsam])
        print("Mean Fst from forward sims = {}".format(projected_jfs.Fst()))
        print("Fst from moments = {}".format(moments_fs.Fst()))
        plot_fs(args, moments_fs, (sum_fs0, sum_fs1),
                (sum_samples_fs0, sum_samples_fs1))
        if args.show2d is True:
            moments.Plotting.plot_2d_comp_Poisson(args.theta * moments_fs,
                                                  projected_jfs, resid_range=10)



"""
Evolution of a single genomic region under Gaussian Stabilizing Selection
with a moving optimum, or "GSSmo".

The simulation allows the recording of quantitative-genetics statistics
over time as well as the preservation of samples into the tree sequence
after the optimum shift.

The demographic model is constant-size Wright-Fisher for 10N generations
around an optimal trait value of zero.  Then, the optimum changes and
the population continues to evolve for a length of time specified
by the user.

The mutation model is uniform across the genome with effect sizes
given by a Gaussian distribution with mean zero and a user-specified
standard deviation.

The genetic model is additive effects with fitness determined by
Gaussian stabilizing selection based on the squared distance from
the optimum.
"""

import fwdpy11
from collections import namedtuple
import pandas as pd
import sqlite3
import math
import numpy as np
import pickle
import lzma
import sys
import os
import argparse

# Simulations with tree sequence recording need
# to know the max position in a genome.  Here,
# we use a length of 1.0. Thus, all mutation
# and recombination events will be uniform
# random variables on the continuous interval
# [0, GENOME_LENGTH).
GENOME_LENGTH = 1.0

# When recording quant-genetic statistics during a simulation,
# we will use this type. Named tuples are extremely efficient,
# and they are easily converted into Pandas DataFrame objects,
# which is very convenient for analysis and output.
DATA = namedtuple('Data', ['generation', 'zbar', 'vg', 'wbar'])


def make_parser():
    """
    Create a command-line interface to the script.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    required = parser.add_argument_group("Required arguments")
    required.add_argument("--popsize", "-N", type=int,
                          help="Diploid population size")
    required.add_argument("--mu", "-m", type=float,
                          help="Mutation rate (per gamete, per generation)")
    required.add_argument("--sigma", "-s", type=float,
                          help="Standard deviation of Gaussian"
                          "distribution of mutational effects")
    required.add_argument("--filename", "-f", type=str, default=None,
                          help="Output file name.  The population will be"
                          "pickled to this file, and the file will be"
                          "compressed using the lzma algorithm")

    optional = parser.add_argument_group("Optional arguments")
    optional.add_argument("--rho", type=float, default=1000.0,
                          help="Scaled recombination rate, rho=4Nr")
    optional.add_argument("--VS", type=float, default=1.0,
                          help="Inverse strength of stabilizing selection")
    optional.add_argument("--opt", type=float, default=1.0,
                          help="Value of new phenotypic optimum")
    optional.add_argument("--time", type=float, default=0.1,
                          help="Amount of time to simulate past"
                          "optimum shift, in units of N")
    optional.add_argument("--record", "-r", action="store_true",
                          help="Record statistics such as VG from population "
                          "each generation")
    optional.add_argument('--statfile', '-S', type=str, default=None,
                          help="File name to record statistics."
                          "The format is an sqlite3 database.")
    optional.add_argument("--preserve", type=int, default=-1,
                          help="Record ancient samples every X generations"
                          "after optimum shift. A value of -1 means not"
                          "to record.")
    optional.add_argument("--num_ind", '-n', type=int, default=50,
                          help="Number of diploids to record as"
                          "ancient samples")
    optional.add_argument("--seed", type=int, default=42,
                          help="Random number seed.")

    return parser


def validate_arguments(args):
    """
    Validate input arguments.
    Note: this is likely incomplete.
    """
    if args.popsize < 0:
        raise ValueError("Population size must be non-negative")
    if args.mu < 0 or math.isfinite(args.mu) is False:
        raise ValueError("Mutation rate must be non-negative and finite")
    if args.sigma < 0 or math.isfinite(args.sigma) is False:
        raise ValueError("Std. dev. of distribution of effect sizes"
                         "must be non-negative and finite")
    if args.filename is None:
        raise ValueError("ouptut filename cannot be None")

    if args.record > 0 and args.statfile is None:
        raise ValueError("Must profile a stats file name when recording"
                         "statistics during simulation")

    if args.statfile is not None and os.path.exists(args.statfile):
        raise RuntimeError("statfile already exists!")

    if args.preserve > 0:
        if args.num_ind > args.popsize:
            raise ValueError("Number of ancient samples is"
                             "greater than population size")


class Recorder(object):
    """
    fwdpy11 allows you to define objects that record data
    from populations during simulation.  Such objects must
    be callable, and the easiest way to do things is to
    create a class with a __call__ function.
    """

    def __init__(self, record_stats, interval, nsam):
        self.data = []
        self.record = record_stats
        self.interval = interval
        self.nsam = nsam

    def __call__(self, pop, recorder):
        if self.record is True:
            # Record mean trait value each generation.
            t = np.array(pop.diploid_metadata, copy=False)
            self.data.append(DATA(pop.generation, t['g'].mean(),
                                  t['g'].var(), t['w'].mean()))

        if self.interval > 0 and pop.generation >= 10*pop.N:
            if pop.generation % self.interval == 0.0:
                if self.nsam < pop.N:
                    s = np.random.choice(pop.N, self.nsam, replace=False)
                else:
                    s = np.arange(pop.N, dtype=np.int32)

                recorder.assign(s)


def runsim(args):
    """
    Run the simulation and deliver output to files.
    """
    pop = fwdpy11.DiploidPopulation(args.popsize, GENOME_LENGTH)

    np.random.seed(args.seed)
    rng = fwdpy11.GSLrng(args.seed)

    GSSmo = fwdpy11.GSSmo(
        [(0, 0, args.VS), (10*args.popsize, args.opt, args.VS)])

    popsizes = [args.popsize]*(10*args.popsize +
                               int(args.time*float(args.popsize)))
    p = {'nregions': [],  # No neutral mutations -- add them later!
         'gvalue': fwdpy11.Additive(2.0, GSSmo),
         'sregions': [fwdpy11.GaussianS(0, GENOME_LENGTH, 1, args.sigma)],
         'recregions': [fwdpy11.Region(0, GENOME_LENGTH, 1)],
         'rates': (0.0, args.mu, args.rho/float(4*args.popsize)),
         # Keep mutations at frequency 1 in the pop if they affect fitness.
         'prune_selected': False,
         'demography':  np.array(popsizes, dtype=np.uint32)
         }
    params = fwdpy11.ModelParams(**p)

    r = Recorder(args.record, args.preserve, args.num_ind)
    fwdpy11.evolvets(rng, pop, params, 100, r, suppress_table_indexing=True)

    with lzma.open(args.filename, 'wb') as f:
        pickle.dump(pop, f)

    if args.statfile is not None:
        stats = pd.DataFrame(r.data, columns=DATA._fields)
        # Write the statistics to an sqlite3 database,
        # which can be processed in R via dplyr.
        conn = sqlite3.connect(args.statfile)
        stats.to_sql('data', conn)


if __name__ == "__main__":
    parser = make_parser()
    args = parser.parse_args(sys.argv[1:])
    validate_arguments(args)
    runsim(args)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

# This file tests that types in population objects
# are indeed "opaque" and not being copied to/from
# Python lists

import unittest
import fwdpy11


class testDiploidPopulation(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.pop = fwdpy11.DiploidPopulation(1000)

    def test_mutations(self):
        self.assertTrue(
            type(self.pop.mutations)
            is fwdpy11.MutationVector)
        self.assertFalse(type(self.pop.mutations) is list)
        self.assertTrue(type(list(self.pop.mutations)) is list)

    def test_diploids(self):
        self.assertTrue(
            type(self.pop.diploids)
            is fwdpy11.DiploidVector)

    def test_a_diploid(self):
        self.assertTrue(
            type(self.pop.diploids[0])
            is fwdpy11.DiploidGenotype)

    def test_gametes(self):
        self.assertTrue(
            type(self.pop.haploid_genomes)
            is fwdpy11.HaploidGenomeVector)


class DiploidTypeSampler(object):
    def __call__(self, pop):
        assert(type(pop.mutations) is fwdpy11.MutationVector)
        assert(type(pop.mutations) is not list)
        assert(type(list(pop.mutations)) is list)
        assert(type(pop.diploids) is fwdpy11.DiploidVector)
        assert(type(pop.diploids[0])
               is fwdpy11.DiploidGenotype)
        assert(type(pop.haploid_genomes) is fwdpy11.HaploidGenomeVector)


class testDiploidPopulationSampler(unittest.TestCase):
    @classmethod
    def setUp(self):
        from fwdpy11.ezparams import mslike
        from fwdpy11 import Multiplicative
        from fwdpy11 import ModelParams
        self.sampler = DiploidTypeSampler()
        self.pop = fwdpy11.DiploidPopulation(1000)
        self.params_dict = mslike(self.pop, simlen=10)
        self.params_dict['gvalue'] = Multiplicative(2.)
        self.params = ModelParams(**self.params_dict)
        self.rng = fwdpy11.GSLrng(42)

    def testSampler(self):
        from fwdpy11 import evolve_genomes as evolve
        try:
            evolve(self.rng, self.pop, self.params, self.sampler)
        except:
            self.fail("unexpected AssertionError")


if __name__ == "__main__":
    unittest.main()



def exponential_growth_rate(N0, Nt, time):
    """
    :param N0: Population size at onset of size change
    :param Nt: Population size to reach at end of size change
    :param time: Time (in generations) to get from N0 to Nt

    :return: A list of integers representing population size over time.

    .. versionadded:: 0.6.0
    """
    if time < 1:
        raise ValueError("time must be >= 1")
    if N0 < 1 or Nt < 1:
        raise ValueError("N0 and Nt must both be >= 1")
    import numpy as np
    return np.exp((np.log(Nt) - np.log(N0))/time)


def migration_matrix_single_extant_deme(ndemes, focal_deme):
    """
    Sets up a migration matrix for a system of ndemes
    but where only one deme has a size > 0.  This
    is a convenience function for simulations where
    demes will appear later on and have migration amongst them.

    :param ndemes: The number of demes that will eventually exist.
    :param focal_deme: The index of the initial extant deme.
    :rtype: numpy.ndarray

    .. versionadded:: 0.6.0
    """
    import numpy as np
    m = np.zeros(ndemes*ndemes).reshape(ndemes, ndemes)
    m[focal_deme, focal_deme] = 1.0
    return m



# Unit tests for fwdpy11.regions

import fwdpy11
import numpy as np
import unittest
import pickle

# Choose global values that differ from default values in Python class constructors
BEG, END, WEIGHT, DOM, LABEL, COUPLED = 0.0, 1.0, 0.5, 0.5, 63, False


class testRegion(unittest.TestCase):
    """
    These tests all apply to Sregions b/c an
    Sregion contains a Region
    """

    def testBadBoundaries(self):
        with self.assertRaises(ValueError):
            fwdpy11.Region(np.nan, 1, 1.0)
        with self.assertRaises(ValueError):
            fwdpy11.Region(0, np.nan, 1.0)
        with self.assertRaises(ValueError):
            fwdpy11.Region(0, 0, 1.0)
        with self.assertRaises(ValueError):
            fwdpy11.Region(0, -1, 1.0)

    def testBadWeight(self):
        with self.assertRaises(ValueError):
            fwdpy11.Region(0, 1, np.nan)
        with self.assertRaises(ValueError):
            fwdpy11.Region(0, 1, -1e-3)


class test_PickleRegion(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.Region(BEG, END, WEIGHT, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.b, up.b)
        self.assertEqual(self.r.e, up.e)
        self.assertEqual(self.r.w, up.w)
        self.assertEqual(self.r.c, up.c)
        self.assertEqual(self.r.l, up.l)


class testExpS(unittest.TestCase):
    # NOTE: this tests applies to all Sregions
    # b/c the scaling param is part of the base class
    def testBadScaling(self):
        with self.assertRaises(ValueError):
            fwdpy11.ExpS(0, 1, 1.0, -0.1, scaling=np.nan)

    def testBadMean(self):
        with self.assertRaises(ValueError):
            fwdpy11.ExpS(0, 1, 1, np.nan)

    def testBadDominance(self):
        with self.assertRaises(ValueError):
            fwdpy11.ExpS(0, 1, 1, -1.0, h=np.nan)


class test_PickleExpS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.ExpS(BEG, END, WEIGHT, -0.2, DOM, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.b, up.b)
        self.assertEqual(self.r.e, up.e)
        self.assertEqual(self.r.w, up.w)
        self.assertEqual(self.r.c, up.c)
        self.assertEqual(self.r.l, up.l)
        self.assertEqual(up.mean, -0.2)
        self.assertEqual(up.h, DOM)


class testGammaS(unittest.TestCase):
    def testBadMean(self):
        with self.assertRaises(ValueError):
            fwdpy11.GammaS(0, 1, 1, mean=np.nan, shape=1.0)

    def testBadShape(self):
        with self.assertRaises(ValueError):
            fwdpy11.GammaS(0, 1, 1, mean=1., shape=np.nan)

    def testBadDominance(self):
        with self.assertRaises(ValueError):
            fwdpy11.GammaS(0, 1, 1, mean=1., shape=1., h=np.nan)


class test_PickleGammaS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.GammaS(BEG, END, WEIGHT,
                                -0.2, 0.5, DOM, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.b, up.b)
        self.assertEqual(self.r.e, up.e)
        self.assertEqual(self.r.w, up.w)
        self.assertEqual(self.r.c, up.c)
        self.assertEqual(self.r.l, up.l)
        self.assertEqual(up.mean, -0.2)
        self.assertEqual(up.shape, 0.5)
        self.assertEqual(up.h, DOM)


class testUniformS(unittest.TestCase):
    def testBadRange(self):
        with self.assertRaises(ValueError):
            fwdpy11.UniformS(0, 1, 1, 0, 0)
        with self.assertRaises(ValueError):
            fwdpy11.UniformS(0, 1, 1, np.nan, 1)
        with self.assertRaises(ValueError):
            fwdpy11.UniformS(0, 1, 1, 1, np.nan)

    def testBadDominance(self):
        with self.assertRaises(ValueError):
            fwdpy11.UniformS(0, 1, 1, lo=1., hi=1., h=np.nan)


class test_PickleUniformS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.UniformS(BEG, END, WEIGHT,
                                  -0.2, 0.5, DOM, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.b, up.b)
        self.assertEqual(self.r.e, up.e)
        self.assertEqual(self.r.w, up.w)
        self.assertEqual(self.r.c, up.c)
        self.assertEqual(self.r.l, up.l)
        self.assertEqual(up.lo, -0.2)
        self.assertEqual(up.hi, 0.5)
        self.assertEqual(up.h, DOM)


class testGaussianS(unittest.TestCase):
    def testBadStdDev(self):
        with self.assertRaises(ValueError):
            fwdpy11.GaussianS(0, 1, 1, np.nan)
        with self.assertRaises(ValueError):
            fwdpy11.GaussianS(0, 1, 1, 0)
        with self.assertRaises(ValueError):
            fwdpy11.GaussianS(0, 1, 1, -1e-3)

    def testBadDominance(self):
        with self.assertRaises(ValueError):
            fwdpy11.GaussianS(0, 1, 1, 1e-3, h=np.nan)


class test_PickleGaussianS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.GaussianS(BEG, END, WEIGHT,
                                   0.5, DOM, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.b, up.b)
        self.assertEqual(self.r.e, up.e)
        self.assertEqual(self.r.w, up.w)
        self.assertEqual(self.r.c, up.c)
        self.assertEqual(self.r.l, up.l)
        self.assertEqual(up.sd, 0.5)
        self.assertEqual(up.h, DOM)


class testConstantS(unittest.TestCase):
    def testBadValue(self):
        with self.assertRaises(ValueError):
            fwdpy11.ConstantS(0, 1, 1, np.nan)

    def testBadDominance(self):
        with self.assertRaises(ValueError):
            fwdpy11.ConstantS(0, 1, 1, 1, np.nan)


class test_PickleConstantS(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.ConstantS(BEG, END, WEIGHT,
                                   0.5, DOM, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r.b, up.b)
        self.assertEqual(self.r.e, up.e)
        self.assertEqual(self.r.w, up.w)
        self.assertEqual(self.r.c, up.c)
        self.assertEqual(self.r.l, up.l)
        self.assertEqual(up.s, 0.5)
        self.assertEqual(up.h, DOM)


class testMutationRegions(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.nregions = [fwdpy11.Region(0, 1, 1)]
        self.sregions = [fwdpy11.ConstantS(
            0, 1, 1, -0.1, 0.25), fwdpy11.ExpS(0, 1, 0.001, 0.01, 1.0)]

    def test_create(self):
        mr = fwdpy11.MutationRegions.create(0.5, self.nregions, self.sregions)  # NOQA


class testMultivariateGaussianEffects(unittest.TestCase):
    def testConstruction(self):
        try:
            fwdpy11.MultivariateGaussianEffects(
                0, 1, 1, np.identity(2))
        except Exception:
            self.fail("Unexpected exception during object construction")

    def testInvalidMatrixShape(self):
        # Input matrix has wrong shape:
        m = np.array([-1.0] * 4)
        with self.assertRaises(ValueError):
            fwdpy11.MultivariateGaussianEffects(
                0, 1, 1, m)

    def testInvalidMatrixShape2(self):
        m = np.array([-1.0] * 6).reshape((3, 2))
        with self.assertRaises(ValueError):
            fwdpy11.MultivariateGaussianEffects(
                0, 1, 1, m)

    def testInvalidMatrix(self):
        # This is not a positive-definite covariance
        # matrix:
        m = np.array([-1.0] * 4).reshape((2, 2))
        with self.assertRaises(ValueError):
            fwdpy11.MultivariateGaussianEffects(
                0, 1, 1, m)

    def testMatrixWithNAN(self):
        m = np.identity(4).reshape((4, 4))
        m[0, 1] = np.nan
        with self.assertRaises(ValueError):
            fwdpy11.MultivariateGaussianEffects(0, 1, 1, m)


class test_PickleMultivariateGaussianEffects(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.r = fwdpy11.MultivariateGaussianEffects(
            BEG, END, WEIGHT, np.identity(2), -0.3, DOM, COUPLED, LABEL)

    def test_pickling(self):
        p = pickle.dumps(self.r, -1)
        up = pickle.loads(p)
        self.assertEqual(self.r, up)


class testRecombinationRegions(unittest.TestCase):
    def test_create(self):
        rr = fwdpy11.RecombinationRegions(1e-3,
                                          [fwdpy11.Region(0, 1, 1),
                                              fwdpy11.Region(1, 2, 1)])
        self.assertEqual(len(rr.weights), 2)

    def test_bad_input(self):
        with self.assertRaises(TypeError):
            fwdpy11.RecombinationRegions(1e-3,
                                         [fwdpy11.ExpS(0, 1, 1, -0.2),
                                          fwdpy11.GaussianS(1, 2, 1, 0.25)])


class testPoissonInterval(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pi = fwdpy11.PoissonInterval(0, 1, 1e-3)

    def test_pickling(self):
        p = pickle.dumps(self.pi)
        up = pickle.loads(p)
        self.assertEqual(up.beg, self.pi.beg)
        self.assertEqual(up.end, self.pi.end)
        self.assertEqual(up.mean, self.pi.mean)


class testBinomialInterval(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.bi = fwdpy11.BinomialInterval(0, 1, 1e-3)

    def test_pickling(self):
        p = pickle.dumps(self.bi)
        up = pickle.loads(p)
        self.assertEqual(up.beg, self.bi.beg)
        self.assertEqual(up.end, self.bi.end)
        self.assertEqual(up.probability, self.bi.probability)


class testFixedCrossovers(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pi = fwdpy11.FixedCrossovers(0, 1, 10)

    def test_pickling(self):
        p = pickle.dumps(self.pi)
        up = pickle.loads(p)
        self.assertEqual(up.beg, self.pi.beg)
        self.assertEqual(up.end, self.pi.end)
        self.assertEqual(up.nxovers, self.pi.nxovers)


class testPoissonPoint(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pi = fwdpy11.PoissonPoint(1., 0.5)

    def test_pickling(self):
        p = pickle.dumps(self.pi)
        up = pickle.loads(p)
        self.assertEqual(up.position, self.pi.position)
        self.assertEqual(up.mean, self.pi.mean)


class testBinomialPoint(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pi = fwdpy11.BinomialPoint(1., 0.5)

    def test_pickling(self):
        p = pickle.dumps(self.pi)
        up = pickle.loads(p)
        self.assertEqual(up.position, self.pi.position)
        self.assertEqual(up.probability, self.pi.probability)

# class test_Region_repr(unittest.TestCase):
#     @classmethod
#     def setUpClass(self):
#         self.nregion = fwdpy11.Region(0, 1, 1)
#         self.constants = fwdpy11.ConstantS(0, 1, 1, -0.1, 2.0)
#         self.uniforms = fwdpy11.UniformS(0, 1, 1, 1, 2, 0)
#         self.exps = fwdpy11.ExpS(0, 1, 1, 0.25)
#         self.gaussians = fwdpy11.GaussianS(0, 1, 1, 0.25)
#         self.gammas = fwdpy11.GammaS(0, 1, 1, 0.124, 2.0)
#
#     def test_nregion(self):
#         x = eval('fwdpy11.' + repr(self.nregion))
#
#     def test_constants(self):
#         x = eval('fwdpy11.' + repr(self.constants))
#
#     def test_uniforms(self):
#         x = eval('fwdpy11.' + repr(self.uniforms))
#
#     def test_exps(self):
#         x = eval('fwdpy11.' + repr(self.exps))
#
#     def test_gaussians(self):
#         x = eval('fwdpy11.' + repr(self.gaussians))
#
#     def test_gammas(self):
#         x = eval('fwdpy11.' + repr(self.gammas))
#
#
# class test_Scaling(unittest.TestCase):
#     def test_ConstantSwithoutScaling(self):
#         """
#         Here, the scaling is 1, and the DFE is constant.
#         Thus, all "s" values must be equal to 10.0
#         """
#         x = fwdpy11.ConstantS(0, 1, 1, s=10, scaling=1)
#         c = x.callback()
#         rng = fwdpy11.GSLrng(42)
#         sh = c(rng)
#         self.assertEqual(sh[0], 10.0)
#
#     def test_ConstantSwithScaling(self):
#         """
#         Here, the scaling is 100, and the DFE is constant.
#         Thus, all "s" values must be equal to 10.0/100.0,
#         or 0.1.
#         """
#         x = fwdpy11.ConstantS(0, 1, 1, s=10, scaling=100)
#         c = x.callback()
#         rng = fwdpy11.GSLrng(42)
#         sh = c(rng)
#         self.assertEqual(sh[0], 0.1)


if __name__ == "__main__":
    unittest.main()



import sqlite3

import pandas as pd

with sqlite3.connect("stats.sqlite3") as conn:
    df = pd.read_sql("select * from data", conn)
with sqlite3.connect("master/stats.sqlite3") as conn:
    df2 = pd.read_sql("select * from data", conn)

print(df.equals(df2))



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#


def evolve_genomes(rng, pop, params, recorder=None):
    """
    Evolve a population without tree sequence recordings.  In other words,
    complete genomes must be simulated and tracked.

    :param rng: random number generator
    :type rng: :class:`fwdpy11.GSLrng`
    :param pop: A population
    :type pop: :class:`fwdpy11.DiploidPopulation`
    :param params: simulation parameters
    :type params: :class:`fwdpy11.ModelParams`
    :param recorder: (None) A temporal sampler/data recorder.
    :type recorder: callable

    .. note::
        If recorder is None,
        then :class:`fwdpy11.RecordNothing` will be used.

    """
    import warnings
    # warnings.simplefilter("default")
    warnings.warn(
        "Simulation without tree sequences is being considered for deprecation!",
        PendingDeprecationWarning
    )
    # Test parameters while suppressing warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # Will throw exception if anything is wrong:
        params.validate()

    from ._fwdpy11 import MutationRegions
    from ._fwdpy11 import evolve_without_tree_sequences
    from ._fwdpy11 import dispatch_create_GeneticMap
    pneutral = 0.0
    if params.mutrate_n + params.mutrate_s > 0.0:
        pneutral = params.mutrate_n/(params.mutrate_n+params.mutrate_s)
    mm = MutationRegions.create(pneutral, params.nregions, params.sregions)
    rm = dispatch_create_GeneticMap(params.recrate, params.recregions)

    if recorder is None:
        from ._fwdpy11 import RecordNothing
        recorder = RecordNothing()

    # FIXME: passing params.popsizes is a hack introduced in 0.6.0
    # top maintain API compatibility
    evolve_without_tree_sequences(rng, pop, params.popsizes,
                                  params.mutrate_n, params.mutrate_s,
                                  mm, rm, params.gvalue,
                                  recorder, params.pself, params.prune_selected)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpye1.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

# This is part of fwdpy11's test suite.
# This file contains functions for quickly
# obtaining simulated populations


def quick_neutral_slocus(N=1000, simlen=100):
    from fwdpy11.ezparams import mslike
    from fwdpy11 import ModelParams
    from fwdpy11 import DiploidPopulation, GSLrng
    from fwdpy11 import Multiplicative
    from fwdpy11 import evolve_genomes
    pop = DiploidPopulation(N)
    params_dict = mslike(pop, simlen=simlen)
    params_dict['gvalue'] = Multiplicative(2.)
    params = ModelParams(**params_dict)
    rng = GSLrng(42)
    evolve_genomes(rng, pop, params)
    return pop


def quick_nonneutral_slocus(N=1000, simlen=100, dfe=None):
    from fwdpy11.ezparams import mslike
    from fwdpy11 import ModelParams
    from fwdpy11 import DiploidPopulation, GSLrng
    from fwdpy11 import evolve_genomes
    from fwdpy11 import ExpS
    from fwdpy11 import Multiplicative
    pop = DiploidPopulation(N)
    if dfe is None:
        dfe = ExpS(0, 1, 1, -0.1)
    params_dict = mslike(
        pop, simlen=simlen, dfe=dfe,
        pneutral=0.95)
    params_dict['gvalue'] = Multiplicative(2.0)
    params = ModelParams(**params_dict)
    rng = GSLrng(42)
    evolve_genomes(rng, pop, params)
    return pop


def quick_slocus_qtrait_pop_params(N=1000, simlen=100):
    from fwdpy11 import DiploidPopulation
    from fwdpy11 import GSS
    from fwdpy11 import Additive
    from fwdpy11 import GaussianS, Region
    import numpy as np
    p = {'nregions': [],
         'sregions': [GaussianS(0, 1, 1, 0.25)],
         'recregions': [Region(0, 1, 1)],
         'rates': (0.0, 2e-3, 1e-3),
         'demography': np.array([N] * simlen, dtype=np.uint32),
         'gvalue': Additive(2.0, GSS(VS=1.0, opt=0.0)),
         'prune_selected': False
         }
    pop = DiploidPopulation(N)
    return (pop, p)



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

# This test is motivated by GitHub issue #432.
# When a DES/DFE returns an effect size of zero,
# such variants were not put into the 'smutations'
# field of a genome, due to the way the fwdpp mutation
# base class constructor was being called.  In PR #433,
# we now require a DFE to tell the Mutation type if the
# variant is neutral or not, and the expectation is that
# the answer is "False".  This test creates a new Sregion
# type where all variants have effect size zero and we run
# a quick sim to make sure that they are all in both genomes
# and in the mutation table.

import fwdpy11
import unittest
import EsizeZero


class Recorder(object):
    def __init__(self):
        self.data = []

    def __call__(self, pop, sampler):
        in_genome = [False]*len(pop.mutations)
        for i in pop.diploids:
            for g in [i.first, i.second]:
                for k in pop.haploid_genomes[g].smutations:
                    in_genome[k] = True
        temp = [in_genome[i.key] for i in pop.tables.mutations]
        self.data.append(all([i is True for i in temp]) is True)


class TestEsizeZero(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pdict = {'nregions': [],
                 'sregions': [EsizeZero.EsizeZero(0, 1, 1, True, 0)],
                 'recregions': [fwdpy11.PoissonInterval(0, 1, 1)],
                 'rates': (0, 5, None),
                 'gvalue': fwdpy11.Multiplicative(2.),
                 'demography': fwdpy11.DiscreteDemography(),
                 'simlen': 10
                 }
        self.params = fwdpy11.ModelParams(**pdict)
        self.pop = fwdpy11.DiploidPopulation(500, 1.)
        self.rng = fwdpy11.GSLrng(512035)

    def test_simulation(self):
        r = Recorder()
        fwdpy11.evolvets(self.rng, self.pop, self.params, 1, r)
        self.assertTrue(all([i is True for i in r.data]))


if __name__ == "__main__":
    unittest.main()



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#
import unittest
import fwdpy11
import numpy as np
from collections import namedtuple


def validate_alive_node_metadata(pop):
    nodes = np.array(pop.tables.nodes, copy=False)
    for md in pop.diploid_metadata:
        for i in [0, 1]:
            if md.deme != nodes['deme'][md.nodes[i]]:
                return False
    return True


IndividualLocation = namedtuple(
    "IndividualLocation", ['generation', 'label', 'deme'])


class WhoWhereWhen(object):
    def __init__(self):
        self.data = []

    def __call__(self, pop, sampler):
        for i, j in enumerate(pop.diploid_metadata):
            if i != j.label:
                raise RuntimeError("index does not equal metadata label value")
            self.data.append(IndividualLocation(pop.generation, i, j.deme))


class TestConstantPopulationSize(unittest.TestCase):
    """
    This tests the absence of demographic events.
    """
    @classmethod
    def setUp(self):
        from test_demographic_models import setup_pop_rng
        from test_demographic_models import setup_pdict
        d = fwdpy11.DiscreteDemography()
        self.pop, self.rng = setup_pop_rng()
        self.pdict = setup_pdict(d, 10)
        self.params = fwdpy11.ModelParams(**self.pdict)

    def test_run_sim(self):
        fwdpy11.evolvets(self.rng, self.pop, self.params, 10)
        N = self.pop.N
        self.assertEqual(self.pop.generation, self.params.simlen)
        self.assertEqual(N, self.pop.N)


class TestSimpleMovesAndCopies(unittest.TestCase):
    """
    Basically doing the same tests
    as test_discrete_demography.TestDiscreteDemography
    """
    @classmethod
    def setUp(self):
        self.rng = fwdpy11.GSLrng(644611)
        self.pop = fwdpy11.DiploidPopulation(100, 1.)
        self.pdict = {'nregions': [],
                      'sregions': [],
                      'recregions': [],
                      'rates': (0, 0, 0),
                      'demography': None,
                      'simlen': 1,
                      'gvalue': fwdpy11.Additive(2.)
                      }

    def test_simple_moves_from_single_deme(self):
        d = fwdpy11.DiscreteDemography(
            [fwdpy11.move_individuals(1, 0, 1, 0.5)])
        self.pdict['demography'] = d
        self.pdict['simlen'] = 2
        params = fwdpy11.ModelParams(**self.pdict)
        w = WhoWhereWhen()
        fwdpy11.evolvets(self.rng, self.pop, params, 100, w)
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(len(deme_sizes[0]), 2)
        for i in range(len(deme_sizes[0])):
            self.assertEqual(deme_sizes[1][i], 50)
        self.assertTrue(validate_alive_node_metadata(self.pop))

        # Validate the parents (inefficiently...)
        pdata = [i for i in w.data if i.generation == self.pop.generation-1]
        self.assertEqual(len(set(pdata)), self.pop.N)
        parents_deme_0 = set()
        parents_deme_1 = set()

        for md in self.pop.diploid_metadata:
            for p in md.parents:
                pd = [i.label for i in pdata if i.label == p]
                self.assertEqual(len(pd), 1)
                if md.deme == 0 and pd[0] not in parents_deme_0:
                    parents_deme_0.add(pd[0])
                elif md.deme == 1 and pd[0] not in parents_deme_1:
                    parents_deme_1.add(pd[0])
        self.assertEqual(len(parents_deme_0.intersection(parents_deme_1)), 0)
        self.assertEqual(len(parents_deme_1.intersection(parents_deme_0)), 0)

    def test_simple_copies_from_single_deme(self):
        d = fwdpy11.DiscreteDemography(
            [fwdpy11.copy_individuals(1, 0, 1, 0.5)])
        self.pdict['simlen'] = 2
        self.pdict['demography'] = d
        params = fwdpy11.ModelParams(**self.pdict)
        w = WhoWhereWhen()
        fwdpy11.evolvets(self.rng, self.pop, params, 100, w)
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        expected = {0: 100, 1: 50}
        self.assertEqual(deme_sizes, expected)
        self.assertTrue(validate_alive_node_metadata(self.pop))

        # Validate the parents (inefficiently...)
        pdata = [i for i in w.data if i.generation == self.pop.generation-1]
        self.assertEqual(len(set([i.label for i in pdata])), 100)
        parents_deme_0 = set()
        parents_deme_1 = set()

        for md in self.pop.diploid_metadata:
            for p in md.parents:
                pd = [i.label for i in pdata if i.label == p]
                self.assertEqual(len(pd), 1)
                if md.deme == 0 and pd[0] not in parents_deme_0:
                    parents_deme_0.add(pd[0])
                elif md.deme == 1 and pd[0] not in parents_deme_1:
                    parents_deme_1.add(pd[0])
        self.assertGreater(len(parents_deme_0.intersection(parents_deme_1)), 0)
        self.assertGreater(len(parents_deme_1.intersection(parents_deme_0)), 0)

    def test_simple_back_and_forth_move(self):
        d = fwdpy11.DiscreteDemography(
            [fwdpy11.move_individuals(1, 0, 1, 0.5),
             fwdpy11.move_individuals(2, 1, 0, 1.)])

        self.pdict['demography'] = d
        self.pdict['simlen'] = 3
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(len(deme_sizes[0]), 1)
        for i in range(len(deme_sizes[0])):
            self.assertEqual(deme_sizes[1][i], self.pop.N)
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_simple_back_and_forth_copy(self):
        d = fwdpy11.DiscreteDemography(
            [fwdpy11.copy_individuals(1, 0, 1, 0.5),
             fwdpy11.copy_individuals(2, 1, 0, 1.)])

        self.pdict['demography'] = d
        self.pdict['simlen'] = 3
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        expected = {0: 150, 1: 50}
        self.assertEqual(deme_sizes, expected)
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_simple_moves_from_multiple_demes(self):
        # Set 1/2 the population to start in deme 1:
        self.pop = fwdpy11.DiploidPopulation([50, 50], 1.)
        # In generation 0, we move 1/2 of deme 1 to
        # deme 2, which creates a new deme:
        d = fwdpy11.DiscreteDemography(
            mass_migrations=[fwdpy11.move_individuals(0, 1, 2, 0.5)])
        self.pdict['demography'] = d
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        # We now expect 50, 25, and 25 individuals in
        # demes 0, 1, and 2
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(len(deme_sizes[0]), 3)
        expected = {0: 50, 1: 25, 2: 25}
        for i in range(len(deme_sizes[0])):
            self.assertEqual(deme_sizes[1][i], expected[i])
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_simple_copies_from_multiple_demes(self):
        # Set 1/2 the population to start in deme 1:
        self.pop = fwdpy11.DiploidPopulation([50, 50], 1.)
        # In generation 0, we move 1/2 of deme 1 to
        # deme 2, which creates a new deme:
        d = fwdpy11.DiscreteDemography(
            [fwdpy11.copy_individuals(0, 1, 2, 0.5)])
        self.pdict['demography'] = d
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        # We now expect 50, 50, and 25 individuals in
        # demes 0, 1, and 2
        deme_sizes = self.pop.deme_sizes()
        expected = {0: 50, 1: 50, 2: 25}
        self.assertEqual(len(deme_sizes[0]), len(expected))
        for i in range(len(deme_sizes[0])):
            self.assertEqual(deme_sizes[1][i], expected[i])
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_single_deme_growth(self):
        N1 = 3412
        t = 111
        G = np.exp((np.log(N1)-np.log(self.pop.N))/t)
        g = [fwdpy11.SetExponentialGrowth(16, 0, G)]
        d = fwdpy11.DiscreteDemography(set_growth_rates=g)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 15+t+1
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        self.assertEqual(self.pop.N, N1)
        self.assertEqual(len(self.pop.diploid_metadata), N1)
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_two_deme_growth(self):
        N0 = [90, 10]
        self.pop = fwdpy11.DiploidPopulation(N0, 1.)
        t = [14, 23]  # generations of growth in each deme
        N1 = [5361, 616]
        G0 = np.exp((np.log(N1[0])-np.log(N0[0]))/t[0])
        G1 = np.exp((np.log(N1[1])-np.log(N0[1]))/t[1])

        g = []
        g.append(fwdpy11.SetExponentialGrowth(7, 0, G0))
        g.append(fwdpy11.SetExponentialGrowth(7+t[0], 0, fwdpy11.NOGROWTH))
        g.append(fwdpy11.SetExponentialGrowth(33, 1, G1))
        g.append(fwdpy11.SetExponentialGrowth(33+t[1], 1, fwdpy11.NOGROWTH))
        d = fwdpy11.DiscreteDemography(set_growth_rates=g)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 100
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)

        md = np.array(self.pop.diploid_metadata, copy=False)
        self.assertEqual(self.pop.N, sum(N1))
        deme_sizes = self.pop.deme_sizes()
        for i, j in zip(deme_sizes[1], N1):
            self.assertEqual(i, j)
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_two_deme_growth_with_hard_reset(self):
        N0 = [90, 10]
        self.pop = fwdpy11.DiploidPopulation(N0, 1.)
        t = [14, 23]  # generations of growth in each deme
        N1 = [5361, 616]
        G0 = np.exp((np.log(N1[0])-np.log(N0[0]))/t[0])
        G1 = np.exp((np.log(N1[1])-np.log(N0[1]))/t[1])

        g = []
        g.append(fwdpy11.SetExponentialGrowth(7, 0, G0))
        g.append(fwdpy11.SetExponentialGrowth(33, 1, G1))
        g.append(fwdpy11.SetExponentialGrowth(33+t[1], 1, fwdpy11.NOGROWTH))
        # Cut off the growth in deme 0 after a few generations,
        # and manually set the new deme size to 100 w/no growth
        p = [fwdpy11.SetDemeSize(11, 0, 100)]
        d = fwdpy11.DiscreteDemography(set_deme_sizes=p,
                                       set_growth_rates=g)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 100
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)

        deme_sizes = self.pop.deme_sizes()
        N1 = [100, N1[1]]
        self.assertEqual(self.pop.N, sum(N1))
        for i, j in zip(deme_sizes[1], N1):
            self.assertEqual(i, j)
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_two_deme_growth_without_hard_reset(self):
        N0 = [90, 10]
        self.pop = fwdpy11.DiploidPopulation(N0, 1.)
        t = [14, 23]  # generations of growth in each deme
        N1 = [5361, 616]
        G0 = np.exp((np.log(N1[0])-np.log(N0[0]))/t[0])
        G1 = np.exp((np.log(N1[1])-np.log(N0[1]))/t[1])

        g = []
        g.append(fwdpy11.SetExponentialGrowth(7, 0, G0))
        g.append(fwdpy11.SetExponentialGrowth(7+t[0], 0, fwdpy11.NOGROWTH))
        g.append(fwdpy11.SetExponentialGrowth(33, 1, G1))
        g.append(fwdpy11.SetExponentialGrowth(33+t[1], 1, fwdpy11.NOGROWTH))
        # after X generations of growth, N[0] changes to 100
        # and the growth rate is not reset.
        p = [fwdpy11.SetDemeSize(11, 0, 100, False)]
        d = fwdpy11.DiscreteDemography(set_growth_rates=g,
                                       set_deme_sizes=p)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 100
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)

        md = np.array(self.pop.diploid_metadata, copy=False)
        deme_sizes = self.pop.deme_sizes()
        N1[0] = np.round(100.*np.power(G0, 7 + t[0] - 11))
        self.assertEqual(self.pop.N, sum(N1))
        for i, j in zip(deme_sizes[1], N1):
            self.assertEqual(i, j)
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_two_moves_in_same_generation(self):
        """
        In generation 0, deme 0 splits equally
        into demes 1 and 2.  This should leave
        deme 0 empty and the sizes of demes
        1 and 2 both equal to 0.5 the initial
        size.
        """
        m = [fwdpy11.move_individuals(0, 0, 2, 0.5),
             fwdpy11.move_individuals(0, 0, 1, 0.5)]
        d = fwdpy11.DiscreteDemography(m)
        N0 = self.pop.N
        self.pdict['demography'] = d
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        expected = {0: 0, 1: N0//2, 2: N0//2}
        deme_sizes = self.pop.deme_sizes()
        for i, j in zip(deme_sizes[0], deme_sizes[1]):
            self.assertEqual(expected[i], j)
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_copies_happen_before_moves(self):
        """
        This is a more complex test.
        In the same generation:
        1. All individuals are copied from deme 0 to deme 1.
        2. 50% of deme 0 moves to deme 2.
        3. 50% of deme 0 moves to deme 3.

        This, at the end, the size of deme 1 should be
        the initial size and the size of demes 2 and 3
        should be 0.5*initial size
        """
        m = [fwdpy11.move_individuals(0, 0, 3, 0.5),
             fwdpy11.copy_individuals(0, 0, 1, 1),
             fwdpy11.move_individuals(0, 0, 2, 0.5)]
        d = fwdpy11.DiscreteDemography(m)
        N0 = self.pop.N
        self.pdict['demography'] = d
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        expected = {0: 0, 1: N0, 2: N0//2, 3: N0//2}
        deme_sizes = self.pop.deme_sizes()
        for i, j in zip(deme_sizes[0], deme_sizes[1]):
            self.assertEqual(expected[i], j)
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_mass_move_with_growth(self):
        """
        In generation 5, the mass movement from 0 to 1
        will reset the growth rate in deme 0 to
        fwdpy11.NOGROWTH

        This test is also handy b/c growth results in
        an odd total N by the time the mass migration happens
        """
        m = [fwdpy11.move_individuals(5, 0, 1, 0.5)]
        g = [fwdpy11.SetExponentialGrowth(0, 0, 1.1)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m, set_growth_rates=g)
        N0 = self.pop.N
        self.pdict['demography'] = d
        self.pdict['simlen'] = 10
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        deme_sizes = self.pop.deme_sizes()
        N5 = np.round(N0*np.power(g[0].G, 5))
        N_after_mass_mig = [N5//2, N5-N5//2]
        for i, j in zip(deme_sizes[1], N_after_mass_mig):
            self.assertEqual(i, j)
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_mass_move_with_growth_no_reset(self):
        """
        In generation 5, the mass movement from 0 to 1
        """
        m = [fwdpy11.move_individuals(5, 0, 1, 0.5, False)]
        g = [fwdpy11.SetExponentialGrowth(0, 0, 1.1)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m, set_growth_rates=g)
        N0 = self.pop.N
        self.pdict['demography'] = d
        self.pdict['simlen'] = 10
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        deme_sizes = self.pop.deme_sizes()
        N5 = np.round(N0*np.power(g[0].G, 5))
        N_after_mass_mig_0 = np.round((N5//2)*np.power(g[0].G, 5))
        N = [N_after_mass_mig_0, N5 - N5//2]
        for i, j in zip(N, deme_sizes[1]):
            self.assertEqual(i, j)
        self.assertTrue(validate_alive_node_metadata(self.pop))


class TestSimpleDemeSizeChanges(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.rng = fwdpy11.GSLrng(42)
        self.pop = fwdpy11.DiploidPopulation(100, 1.)
        self.pdict = {'nregions': [],
                      'sregions': [],
                      'recregions': [],
                      'rates': (0, 0, 0),
                      'demography': None,
                      'simlen': 1,
                      'gvalue': fwdpy11.Additive(2.)
                      }

    def test_global_extinction_single_deme(self):
        ds = [fwdpy11.SetDemeSize(4, 0, 0)]
        d = fwdpy11.DiscreteDemography(set_deme_sizes=ds)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 10
        params = fwdpy11.ModelParams(**self.pdict)
        with self.assertRaises(fwdpy11.GlobalExtinction):
            fwdpy11.evolvets(self.rng, self.pop, params, 100)

        self.assertEqual(self.pop.N, 100)
        self.assertEqual(self.pop.generation, 4)
        nodes = np.array(self.pop.tables.nodes)
        self.assertTrue(np.all(nodes['time'][self.pop.alive_nodes] == 4))

    def test_no_valid_parents(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 10
        params = fwdpy11.ModelParams(**self.pdict)
        with self.assertRaises(fwdpy11.DemographyError):
            fwdpy11.evolvets(self.rng, self.pop, params, 100)

    def test_no_valid_parents_alternate_method(self):
        m = [fwdpy11.copy_individuals(0, 0, 1, 1.0)]
        s = [fwdpy11.SetDemeSize(0, 0, 0), fwdpy11.SetDemeSize(2, 0, 100)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m, set_deme_sizes=s)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 10
        params = fwdpy11.ModelParams(**self.pdict)
        with self.assertRaises(fwdpy11.DemographyError):
            fwdpy11.evolvets(self.rng, self.pop, params, 100)

    def test_no_valid_parents_with_migration(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5]*4).reshape(2, 2)
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s,
                                       migmatrix=M)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 10
        params = fwdpy11.ModelParams(**self.pdict)
        with self.assertRaises(fwdpy11.DemographyError):
            fwdpy11.evolvets(self.rng, self.pop, params, 100)

    def test_no_valid_parents_with_migration_2(self):
        m = [fwdpy11.move_individuals(0, 0, 1, 1)]
        s = [fwdpy11.SetDemeSize(0, 0, 100)]
        M = np.array([0.5]*4).reshape(2, 2)
        cM = [fwdpy11.SetMigrationRates(
            0, np.array([0, 1, 0, 1]).reshape(2, 2))]
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_deme_sizes=s,
                                       set_migration_rates=cM,
                                       migmatrix=M)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 10
        params = fwdpy11.ModelParams(**self.pdict)
        try:
            fwdpy11.evolvets(self.rng, self.pop, params, 100)
        except Exception:
            self.fail("unexpected exception")


class TestSimpleMigrationModels(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.rng = fwdpy11.GSLrng(42)
        self.pop = fwdpy11.DiploidPopulation(100, 1.)
        self.pdict = {'nregions': [],
                      'sregions': [],
                      'recregions': [],
                      'rates': (0, 0, 0),
                      'demography': None,
                      'simlen': 1,
                      'gvalue': fwdpy11.Additive(2.)
                      }

    def test_simple_two_deme_migration(self):
        mm = np.array([0.5]*4).reshape(2, 2)
        self.pop = fwdpy11.DiploidPopulation([50, 50], 1.)
        d = fwdpy11.DiscreteDemography(migmatrix=mm)
        N = self.pop.N
        self.pdict['demography'] = d
        self.pdict['simlen'] = 5
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        self.assertEqual(N, self.pop.N)
        deme_sizes = self.pop.deme_sizes()
        for i in deme_sizes[1]:
            self.assertEqual(i, self.pop.N//2)
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_change_migration_rates_simple_two_deme_migration(self):
        """
        For a 2-deme model, the mig matrix is
        [[0, 1]
         [1, 0]]
        so that all offspring have both parents from the other deme.

        After 3 generations, we reset the migration rates to be
        [[1, 0],
         [1, 0]]
        so that all parents are from deme zero.
        """
        mm = np.array([0, 1, 1, 0]).reshape(2, 2)
        mmigs = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        smr = [fwdpy11.SetMigrationRates(
            3, np.array([1, 0, 1, 0]).reshape(2, 2))]
        d = fwdpy11.DiscreteDemography(mass_migrations=mmigs, migmatrix=mm,
                                       set_migration_rates=smr)
        N = self.pop.N
        self.pdict['demography'] = d
        self.pdict['simlen'] = 5
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        self.assertEqual(N, self.pop.N)
        deme_sizes = self.pop.deme_sizes()
        for i in deme_sizes[1]:
            self.assertEqual(i, self.pop.N//2)
        self.assertTrue(validate_alive_node_metadata(self.pop))

    def test_change_migration_rates_simple_two_deme_migration_bad_matrix(self):
        """
        For a 2-deme model, the mig matrix is
        [0, 1
         1, 0]
        so that all offspring have both parents from the other deme.

        After 3 generations, we reset the migration rates to be
        [[0.5, 0.5],
         [0, 0]],
        which leads to there being no parents for deme 1, raising a
        fwdpy11.DemographyError exception.
        """
        mm = np.array([0, 1, 1, 0]).reshape(2, 2)
        mmigs = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        smr = [fwdpy11.SetMigrationRates(
            3, np.array([0.5, 0.5, 0, 0]).reshape(2, 2))]
        d = fwdpy11.DiscreteDemography(mass_migrations=mmigs, migmatrix=mm,
                                       set_migration_rates=smr)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 5
        params = fwdpy11.ModelParams(**self.pdict)
        with self.assertRaises(fwdpy11.DemographyError):
            fwdpy11.evolvets(self.rng, self.pop, params, 100)

    def test_selfing_vs_migration(self):
        """
        Parents of deme 0 are all migrants from deme 1.
        Parents of deme 1 are all migrants from deme 0.
        Deme 0 never selfs.  Deme 1 always selfs.

        NOTE: currently doesn't test anything.
        """
        mm = np.array([0, 1, 1, 0]).reshape(2, 2)
        mmigs = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        s = [fwdpy11.SetSelfingRate(0, 1, 1.0)]
        mmigs = [fwdpy11.move_individuals(0, 0, 1, 0.5)]
        d = fwdpy11.DiscreteDemography(
            mass_migrations=mmigs, migmatrix=mm, set_selfing_rates=s)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 5
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        self.assertTrue(validate_alive_node_metadata(self.pop))


class TestGrowthModels(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.rng = fwdpy11.GSLrng(42)
        self.pop = fwdpy11.DiploidPopulation(100, 1.)
        self.pdict = {'nregions': [],
                      'sregions': [],
                      'recregions': [],
                      'rates': (0, 0, 0),
                      'demography': None,
                      'simlen': 1,
                      'gvalue': fwdpy11.Additive(2.)
                      }

    def test_global_extinction_single_deme(self):
        """
        Pop would go exctinct in generation 8,
        which triggers exception in generation 7
        """
        g = [fwdpy11.SetExponentialGrowth(0, 0, 0.5)]
        d = fwdpy11.DiscreteDemography(set_growth_rates=g)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 20
        params = fwdpy11.ModelParams(**self.pdict)
        with self.assertRaises(fwdpy11.GlobalExtinction):
            fwdpy11.evolvets(self.rng, self.pop, params, 100)
        self.assertEqual(self.pop.generation, 7)

        nodes = np.array(self.pop.tables.nodes)
        self.assertTrue(np.all(nodes['time'][self.pop.alive_nodes] == 7))

    def test_shrink_to_zero_then_recolonize(self):
        m = [fwdpy11.copy_individuals(0, 0, 1, 1.0),
             fwdpy11.copy_individuals(10, 1, 0, 1.0)]
        g = [fwdpy11.SetExponentialGrowth(0, 0, 0.5)]
        d = fwdpy11.DiscreteDemography(mass_migrations=m,
                                       set_growth_rates=g)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 20
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 100)
        deme_sizes = self.pop.deme_sizes(as_dict=True)
        self.assertEqual(deme_sizes, {0: 100, 1: 100})


class TestGeneticValueLists(unittest.TestCase):
    """
    Added in 0.6.0 to test exceptions raised
    when the number of genetic value objects
    are out of whack vis-a-vis the demography.
    """
    @classmethod
    def setUp(self):
        self.rng = fwdpy11.GSLrng(42)
        self.pop = fwdpy11.DiploidPopulation(100, 1.)
        self.pdict = {'nregions': [],
                      'sregions': [],
                      'recregions': [],
                      'rates': (0, 0, 0),
                      'demography': None,
                      'simlen': 1,
                      'gvalue': fwdpy11.Additive(2.)
                      }

    def test_too_many(self):
        self.pdict['demography'] = np.array([self.pop.N]*10, dtype=np.uint32)
        self.pdict['gvalue'] = [fwdpy11.Additive(2.0)]*2
        params = fwdpy11.ModelParams(**self.pdict)
        with self.assertRaises(ValueError):
            fwdpy11.evolvets(self.rng, self.pop, params, 100)

    def test_too_few(self):
        mmigs = [fwdpy11.move_individuals(
            0, 0, 1, 1./3.), fwdpy11.move_individuals(0, 0, 2, 1./3.)]
        d = fwdpy11.DiscreteDemography(mass_migrations=mmigs)
        self.pdict['demography'] = d
        self.pdict['simlen'] = 5
        self.pdict['gvalue'] = [fwdpy11.Additive(2.0)]*2  # This is the error
        params = fwdpy11.ModelParams(**self.pdict)
        with self.assertRaises(ValueError):
            fwdpy11.evolvets(self.rng, self.pop, params, 100)

    def test_too_few_alt_method(self):
        self.pop = fwdpy11.DiploidPopulation(
            [0, self.pop.N//3, 2*self.pop.N//3], 1)
        self.pdict['demography'] = np.array([self.pop.N]*100, dtype=np.uint32)
        self.pdict['gvalue'] = [fwdpy11.Additive(2.0)]*2  # This is the error
        params = fwdpy11.ModelParams(**self.pdict)
        with self.assertRaises(ValueError):
            fwdpy11.evolvets(self.rng, self.pop, params, 100)


class TestIMModel(unittest.TestCase):
    """
    Test of an Isolation-with-Migration, or IM model.

    We also test that we can evolve a pop up to a certain
    point, stop, then evolve it "the rest of the way". Such
    a test is useful for simulations of human OOA models
    where we may want to simplify more often during the
    exponential growth phase in order to save memory.
    """
    def getG(N0, Nt, t):
        return np.exp((np.log(Nt) - np.log(N0))/t)

    @classmethod
    def setUp(self):
        self.rng = fwdpy11.GSLrng(666)
        self.Nref = 200
        self.Tsplit = 10
        psplit = 0.2
        self.gens_post_split = 5
        self.N0t = 300
        self.N1t = 400
        G0 = self.getG(self.Nref*(1.-psplit), self.N0t, self.gens_post_split)
        G1 = self.getG(self.Nref*(psplit), self.N1t, self.gens_post_split)

        split = [fwdpy11.move_individuals(self.Tsplit, 0, 1, psplit)]
        growth = [fwdpy11.SetExponentialGrowth(self.Tsplit, 0, G0),
                  fwdpy11.SetExponentialGrowth(self.Tsplit, 1, G1)]
        m = np.zeros(4).reshape(2, 2)
        m[0, 0] = 1
        cm = [fwdpy11.SetMigrationRates(self.Tsplit, 0, [0.98, 0.02]),
              fwdpy11.SetMigrationRates(self.Tsplit, 1, [0.02, 0.98])]
        d = fwdpy11.DiscreteDemography(mass_migrations=split,
                                       set_growth_rates=growth,
                                       set_migration_rates=cm,
                                       migmatrix=m)

        nregions = []
        sregions = []
        recregions = []

        self.pdict = {'nregions': nregions,
                      'sregions': sregions,
                      'recregions': recregions,
                      'rates': (0, 0, 0),
                      'gvalue': fwdpy11.Multiplicative(2.),
                      'demography': d,
                      'simlen': -1,
                      'prune_selected': True
                      }
        self.pop = fwdpy11.DiploidPopulation(self.Nref, 1.0)

    def test_complete_sim(self):
        self.pdict['simlen'] = self.Tsplit + self.gens_post_split
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 5)
        self.assertEqual(self.pop.generation, params.simlen)
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(deme_sizes[1][0], self.N0t)
        self.assertEqual(deme_sizes[1][1], self.N1t)

    def test_evolve_in_two_steps(self):
        self.pdict['simlen'] = self.Tsplit
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 5)
        self.assertEqual(self.pop.generation, self.Tsplit)
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(len(deme_sizes[0]), 1)  # Split hasn't happened
        self.assertEqual(deme_sizes[1][0], self.Nref)

        self.pdict['simlen'] = self.gens_post_split
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 2)  # Simplify more often
        self.assertEqual(self.pop.generation,
                         self.Tsplit + self.gens_post_split)
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(deme_sizes[1][0], self.N0t)
        self.assertEqual(deme_sizes[1][1], self.N1t)

    def test_evolve_in_two_steps_restart_with_two_demes(self):
        """
        Tests the more complex case of restarting a sim
        when multiple demes are present.
        """
        self.pdict['simlen'] = self.Tsplit + 3  # Evolve past the split
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 5)
        self.assertEqual(self.pop.generation, self.Tsplit + 3)
        deme_sizes = self.pop.deme_sizes()
        self.assertTrue(all([i > 0 for i in deme_sizes[1].tolist()]))

        self.pdict['simlen'] = self.gens_post_split - 3
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 2)  # Simplify more often
        self.assertEqual(self.pop.generation,
                         self.Tsplit + self.gens_post_split)
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(deme_sizes[1][0], self.N0t)
        self.assertEqual(deme_sizes[1][1], self.N1t)

    def test_reuse_of_the_demographic_model(self):
        """
        Since DiscreteDemography "remembers" the state
        of the model, we have to test that the model gets
        properly reset when confronted with a "fresh"
        population.
        """
        self.pdict['simlen'] = self.Tsplit + self.gens_post_split
        params = fwdpy11.ModelParams(**self.pdict)
        fwdpy11.evolvets(self.rng, self.pop, params, 5)
        self.assertEqual(self.pop.generation, params.simlen)
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(deme_sizes[1][0], self.N0t)
        self.assertEqual(deme_sizes[1][1], self.N1t)

        npop = fwdpy11.DiploidPopulation(self.Nref, 1.0)
        fwdpy11.evolvets(self.rng, npop, params, 5)
        self.assertEqual(npop.generation, params.simlen)
        deme_sizes = self.pop.deme_sizes()
        self.assertEqual(deme_sizes[1][0], self.N0t)
        self.assertEqual(deme_sizes[1][1], self.N1t)


if __name__ == "__main__":
    unittest.main()



import unittest
import fwdpy11
import pickle
import lzma


class TestOldBinaryFileFormats(unittest.TestCase):
    def test_reading_045(self):
        pop = fwdpy11.DiploidPopulation.load_from_file("tests/v045.bin")
        for i in pop.tables.mutations:
            self.assertEqual(
                pop.mutations[i.key].pos, pop.tables.sites[i.site].position)

    def test_unpickling_045(self):
        with lzma.open('tests/v045.lzma', 'rb') as f:
            pop = pickle.load(f)
        for i in pop.tables.mutations:
            self.assertEqual(
                pop.mutations[i.key].pos, pop.tables.sites[i.site].position)


if __name__ == "__main__":
    unittest.main()



"""
This module provides functions to generate demographic events for
"isolation-with-migration", or IM, models.
"""


def two_deme_IM(Nanc, T, psplit, Ns, migrates, burnin=10.0,
                as_dict=False):
    """
    Isolation-with-migration (IM) model for two demes.

    An ancestral population splits into two daughter demes.
    At the time of the split, ``psplit`` of the ancestral
    population moves into deme 1.  The two daughter populations
    begin exponential growth until the present time and migration
    may occur between them.

    :param Nanc: The ancestral population size.
    :type Nanc: int
    :param T: The time of the split, in units of Nanc generations
              into the past.
    :type T: float
    :param psplit: The proportion of the ancestral population that splits
                   off to found deme 1
    :type psplit: float
    :param Ns: The final sizes of demes 0 and 1, relative to Nanc
    :type Ns: tuple
    :param migrates: The migration rates from 0 to 1 and from 1 to 0,
                     respectively. Migration rates are the fraction
                     of the destination deme replaced by the source
                     deme.
    :type migrates: float
    :param burnin: Time to simulate before the split, in units of Nanc
    :type burnin: float
    :param as_dict: If False, return an instance of
                    :class:`fwdpy11.DiscreteDemography`.
                    Otherwise, return a dict that can be
                    used to construct a ``DiscreteDemography``
                    instance.
    :type as_dict: boolean
    :rtype: tuple
    :returns: the model events, the split time, and
              the time to simulate post-split.

    .. note::

        The events returned by this model assume/require that you will
        construct a population with intitial size ``Nanc``.

    .. versionadded:: 0.6.0
    """
    import fwdpy11
    import numpy as np
    N0, N1 = Ns
    m01, m10 = migrates

    split_time = np.rint(Nanc*burnin).astype(int)
    # The split event
    split = [fwdpy11.move_individuals(when=split_time, source=0,
                                      destination=1,
                                      fraction=psplit)]
    # Get growth rates and set growth rate changes,
    # taking care to handle our rounding!
    gens_post_split = np.rint(Nanc*T).astype(int)
    N0split = np.rint(Nanc*(1.-psplit))
    if N0split == 0 or N0split == Nanc:
        raise ValueError("invalid value for psplit: {}".format(psplit))
    N0final = np.rint(N0*Nanc)
    N1split = np.rint(Nanc*psplit)
    if N1split == 0 or N1split == Nanc:
        raise ValueError("invalid value for psplit: {}".format(psplit))
    N1final = np.rint(N1*Nanc)
    G0 = fwdpy11.exponential_growth_rate(N0split, N0final,
                                         gens_post_split)
    G1 = fwdpy11.exponential_growth_rate(N1split, N1final,
                                         gens_post_split)
    growth = [fwdpy11.SetExponentialGrowth(split_time, 0, G0),
              fwdpy11.SetExponentialGrowth(split_time, 1, G1)]

    # Set up the migration matrix for two demes, but only
    # deme zero exists.
    m = fwdpy11.migration_matrix_single_extant_deme(2, 0)
    # The rows of the matrix change at the split:
    cm = [fwdpy11.SetMigrationRates(split_time, 0, [1.-m10, m10]),
          fwdpy11.SetMigrationRates(split_time, 1, [m01, 1.-m01])]

    mdict = {'mass_migrations': split,
             'set_growth_rates': growth,
             'set_migration_rates': cm,
             'migmatrix': m
             }
    if as_dict is False:
        return fwdpy11.DiscreteDemography(**mdict), split_time, gens_post_split

    return mdict, split_time, gens_post_split



#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import fwdpy11
import sparse
import numpy as np

# Functions related to FS calculation

NOT_A_SAMPLE = np.iinfo(np.int32).min


def _include_both(m):
    return True


def _include_neutral(m):
    return m.neutral is True


def _include_selected(m):
    return m.neutral is False


def _validate_windows(windows, genome_length):
    if windows != sorted(windows, key=lambda x: x[0]):
        raise ValueError("windows must be sorted in increasing order")
    for w in windows:
        if w[0] >= w[1]:
            raise ValueError("windows must be [a, b) and a < b")
        for i in w:
            if i < 0 or i > genome_length:
                raise ValueError(
                    "window coordinates must be [0, genome_length)")
    for i in range(1, len(windows), 2):
        if windows[i-1][0] < windows[i][1] and windows[i][0] < windows[i-1][1]:
            raise ValueError("windows cannot overlap")


def _update_window(windex, tree, windows):
    while windex < len(windows) and tree.right > windows[windex][1]:
        windex += 1
    return windex


def _position_in_window(pos, window):
    return pos >= window[0] and pos < window[1]


def _simplify(tables, samples, simplify):
    if simplify is False:
        return tables, samples

    t, i = fwdpy11.simplify_tables(tables, samples)
    return t, i[samples]


def _1dfs(self, samples, windows, include_function, simplify):
    """
    Returns an array with the zero and fixed
    bins masked out.  The masking is for
    consistency w/ndfs output.
    """
    t, s = _simplify(self, samples, simplify)
    fs = [np.zeros(len(s)+1, dtype=np.int32) for i in windows]
    ti = fwdpy11.TreeIterator(t, s)
    windex = 0
    sites = np.array(t.sites, copy=False)
    positions = sites['position'][:]
    for tree in ti:
        for m in tree.mutations():
            if include_function(m):
                pos = positions[m.site]
                while windex < len(windows) and windows[windex][1] < pos:
                    windex += 1
                if windex >= len(windows):
                    break
                if _position_in_window(pos, windows[windex]):
                    c = np.uint32(tree.leaf_counts(m.node))
                    fs[windex][c] += 1

    rv = []
    for i in fs:
        i = np.ma.array(i)
        i[0] = np.ma.masked
        i[-1] = np.ma.masked
        rv.append(i)
    return rv


def _ndfs(self, samples, sample_groups, num_sample_groups,
          windows, include_function, simplify):
    """
    For more than one sample, always work with the joint FS
    at first. When the marginals are wanted, we extract them
    later.
    """
    shapes = tuple(len(i)+1 for i in samples)
    dok_JFS = [sparse.DOK(shapes, dtype=np.int32) for i in windows]

    sample_list = np.where(sample_groups != NOT_A_SAMPLE)[0]
    t, s = _simplify(self, sample_list, simplify)
    ti = fwdpy11.TreeIterator(t, s, update_samples=True)
    counts = np.zeros(len(samples), dtype=np.int32)
    windex = 0
    sites = np.array(t.sites, copy=False)
    positions = sites['position'][:]
    for tree in ti:
        for m in tree.mutations():
            if include_function(m):
                pos = positions[m.site]
                while windex < len(windows) and windows[windex][1] < pos:
                    windex += 1
                if windex >= len(windows):
                    break
                if _position_in_window(pos, windows[windex]):
                    counts[:] = 0
                    d = tree.samples_below(m.node)
                    if len(d) > 0:
                        c = np.unique(sample_groups[d], return_counts=True)
                        counts[c[0]] += c[1]
                        dok_JFS[windex][tuple((i) for i in counts)] += 1
    return [sparse.COO(i) for i in dok_JFS]


def _handle_fs_marginalizing(fs, marginalize, nwindows, num_sample_groups):
    if marginalize is False or num_sample_groups == 1:
        return fs
    if nwindows == 1:
        temp = {}
        for i in range(num_sample_groups):
            axes = tuple(j for j in range(num_sample_groups) if j != i)
            temp[i] = np.ma.array(fs.sum(axis=axes).todense())
            temp[i][0] = np.ma.masked
            temp[i][-1] = np.ma.masked
        return temp

    rv = []
    for f in fs:
        temp = {}
        for i in range(num_sample_groups):
            axes = tuple(j for j in range(num_sample_groups) if j != i)
            temp[i] = np.ma.array(f.sum(axis=axes).todense())
            temp[i][0] = np.ma.masked
            temp[i][-1] = np.ma.masked
        rv.append(temp)
    return rv


def _fs_implementation(self, samples, windows,
                       include_function, simplify):
    """
    self is a fwdpy11.TableCollection
    samples is a list of 1d numpy.ndarray
    """
    if len(samples) == 0:
        raise ValueError("empty list of samples")

    # Don't figure out sample groups when
    # there is only 1
    if len(samples) == 1:
        return _1dfs(self, samples[0], windows,
                     include_function, simplify)

    sample_groups = np.array([NOT_A_SAMPLE]*len(self.nodes), dtype=np.int32)
    for i, j in enumerate(samples):
        if np.any(sample_groups[j] != NOT_A_SAMPLE):
            raise ValueError(
                "sample nodes cannot be part of multiple sample groups")
        sample_groups[j] = i

    num_sample_groups = i + 1
    return _ndfs(self, samples, sample_groups, num_sample_groups,
                 windows, include_function, simplify)


def _fs(self, samples, marginalize=False,
        simplify=False, windows=None,
        separate_windows=False,
        include_neutral=True,
        include_selected=True):
    """
    :param samples: lists of numpy arrays of sample nodes
    :param marginalize: For ``FS`` involving multiple samples,
                        extract out the marginal ``FS`` per sample.
    :param simplify: If ``True``, simplify with respect to the sample
                     set prior to calculating the ``FS``.
    :param windows: A list of non-overlapping intervals from which
                    the ``FS`` is calculated.
    :param separate_windows: If ``True``, return ``FS`` separately for
                             each interval in ``windows``.
    :param include_neutral: Include neutral mutations?
    :param include_selected: Include selected mutations?

    :returns: The mutation frequency spectrum
    :rtype: object

    The details of the return value depend heavily on the options.

    * If a single sample list is provided, the return value is a
      1-d ndarray with the first and last bins masked.  The unmasked
      bins correspond to frequencies 1 to n-1, where n is the number
      of sample nodes.
    * If multiple sample lists are provided, the return value is a
      ``sparse.COO`` matrix.  For each dimension, the 0 and n bins
      are included.
    * If ``marginalize == True`` and more than one sample bin is provided,
      the ``sparse.COO`` matrix is converted into a dict where the key
      is the index of the sample list and the value is a dense 1-d array,
      masked as described above.
    * If multiple windows are provided and ``separate_windows == False``,
      then the return value is a single frequency spectrum summed over windows.
      If ``separate_windows == True``, then a list of frequency spectra is
      returned, indexed in the same order as the input windows.

    .. versionadded:: 0.6.0

        Python implementation added
    """
    for s in samples:
        if len(s) == 0:
            raise ValueError("samples lists cannot be empty")
        if len(s) < 2:
            raise ValueError("samples lists must have at least two nodes")
        if np.any(s >= len(self.nodes)):
            raise ValueError("invalid samples")

    if windows is None:
        windows = [(0, self.genome_length)]

    _validate_windows(windows, self.genome_length)

    if include_neutral is False and include_selected is False:
        raise ValueError("One or both of include_neutral "
                         "and include_selected must be True")

    include_function = _include_both
    if include_neutral is False:
        include_function = _include_selected
    elif include_selected is False:
        include_function = _include_neutral

    lfs = _fs_implementation(self, samples, windows,
                             include_function, simplify)
    if separate_windows is True:
        lfs = _handle_fs_marginalizing(lfs, marginalize,
                                       len(windows), len(samples))
        return lfs
    fs = lfs[0]
    for i in lfs[1:]:
        fs += i
    fs = _handle_fs_marginalizing(fs, marginalize, len(windows), len(samples))
    return fs


def _patch_table_collection(t):
    t.fs = _fs



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import sys

if sys.version_info[0] < 3:
    raise ValueError("Python3 required!")

from fwdpy11._version import __version__ # NOQA
from ._fwdpy11 import * # NOQA
from ._dev import * # NOQA
from ._types.model_params import * # NOQA
from ._types.demography_debugger import DemographyDebugger # NOQA
from ._evolve_genomes import * # NOQA
from ._evolvets import * # NOQA
from ._monkeypatch import _diploid_population # NOQA
from ._monkeypatch import _table_collection # NOQA
from ._monkeypatch import _data_matrix # NOQA
from ._demography import * # NOQA

# NOTE: some operations that can be implemented efficiently
# in Python are supplied as monkey-patches to the pybind11 classes
_monkeypatch._diploid_population._patch_diploid_population(DiploidPopulation) # NOQA
_monkeypatch._table_collection._patch_table_collection(TableCollection) # NOQA
_monkeypatch._data_matrix._patch_data_matrix(DataMatrix) # NOQA



from setuptools import setup
from setuptools import Extension
from setuptools.command.build_ext import build_ext
import sys
import pybind11
import setuptools
import re
import os
import platform
import glob
import subprocess
from distutils.version import LooseVersion


if sys.version_info[0] < 3:
    raise ValueError("Python 3 is required!")

__version__ = '0.6.4'

if sys.version_info < (3, 3):
    raise RuntimeError("Python >= 3.3 required")

if pybind11.__version__ < '2.4.3':
    raise RuntimeError("pybind11 >= " + '2.4.3' + " required")

if sys.version_info >= (3, 7):
    if pybind11.__version__ < '2.3.0':
        raise RuntimeError(
            "Python 3.7 and newer required pybind11 2.3 or greater")


# clang/llvm is default for OS X builds.
# can over-ride darwin-specific options
# with setup.py --gcc install
if '--gcc' in sys.argv:
    USE_GCC = True
    sys.argv.remove('--gcc')
else:
    USE_GCC = False

if '--weffcpp' in sys.argv:
    USE_WEFFCPP = True
    sys.argv.remove('--weffcpp')
else:
    USE_WEFFCPP = False

if '--debug' in sys.argv:
    DEBUG_MODE = True
    sys.argv.remove('--debug')
else:
    DEBUG_MODE = False

if '--enable-profiling' in sys.argv:
    ENABLE_PROFILING = True
    sys.argv.remove('--enable-profiling')
else:
    ENABLE_PROFILING = False

if '--skip_tests' in sys.argv:
    SKIP_BUILDING_TESTS = True
    sys.argv.remove('--skip_tests')
else:
    SKIP_BUILDING_TESTS = False

if '--disable_lto' in sys.argv:
    SKIP_LTO = True
    sys.argv.remove('--disable_lto')
else:
    SKIP_LTO = False

# if '--assert' in sys.argv:
#     ASSERT_MODE = True
#     sys.argv.remove('--assert')
# else:
#     ASSERT_MODE = False


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(
            self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        cfg = 'Debug' if DEBUG_MODE is True else 'Release'

        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += [
                '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j2']

        if ENABLE_PROFILING is True:
            cmake_args += ['-DENABLE_PROFILING=ON']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())

        if USE_WEFFCPP is True:
            cmake_args.append('-DUSE_WEFFCPP=ON')
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        if SKIP_BUILDING_TESTS is True:
            cmake_args.append('-DBUILD_UNIT_TESTS=OFF')

        if SKIP_LTO is True:
            cmake_args.append('-DDISABLE_LTO=ON')

        subprocess.check_call(['cmake', ext.sourcedir] +
                              cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] +
                              build_args, cwd=self.build_temp)


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


PKGS = ['fwdpy11',
        'fwdpy11.demographic_models',
        'fwdpy11._types',
        'fwdpy11._monkeypatch']

INCLUDES = [
    'fwdpy11/headers',
    'fwdpy11/headers/fwdpp',
    # Path to pybind11 headers
    get_pybind_include(),
    get_pybind_include(user=True),
    os.path.join(sys.prefix, 'include')
]

LIBRARY_DIRS = [
    os.path.join(sys.prefix, 'lib')
]

ext_modules = [
    CMakeExtension('fwdpy11._fwdpy11'),
]


# Figure out the headers we need to install:
generated_package_data = {}
for root, dirnames, filenames in os.walk('fwdpy11/headers'):
    if 'testsuite' not in root and 'examples' not in root and \
       'python_examples' not in root:
        g = glob.glob(root + '/*.hh')
        if len(g) > 0:
            replace = root.replace('/', '.')
            # If there's a header file, we add the directory as a package
            if replace not in PKGS:
                PKGS.append(replace)
            generated_package_data[replace] = ['*.hh']
        g = glob.glob(root + '/*.hpp')
        if len(g) > 0:
            replace = root.replace('/',  '.')
            # If there's a header file, we add the directory as a package
            if replace not in PKGS:
                PKGS.append(replace)
            try:
                if '*.hpp' not in generated_package_data[replace]:
                    generated_package_data[replace].append('*.hpp')
            except:
                generated_package_data[replace] = ['*.hpp']
        g = glob.glob(root + '/*.tcc')
        if len(g) > 0:
            replace = root.replace('/', '.')
            # If there's a template implementation file,
            # we add the directory as a package
            if replace not in PKGS:
                PKGS.append(replace)
            try:
                if '*.tcc' not in generated_package_data[replace]:
                    generated_package_data[replace].append('*.tcc')
            except:
                generated_package_data[replace] = ['*.tcc']

long_desc = open("README.rst").read()

setup(
    name='fwdpy11',
    version=__version__,
    author='Kevin Thornton',
    author_email='krthornt@uci.edu',
    url='http://molpopgen.github.io/fwdpy11',
    classifiers=['Intended Audience :: Science/Research',
                 'Topic :: Scientific/Engineering :: Bio-Informatics',
                 'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)'],
    description='Forward-time population genetic simulation in Python',
    license='GPL >= 3',
    requires=['pybind11', 'numpy', 'tskit'],
    provides=['fwdpy11'],
    obsoletes=['none'],
    data_files=[('fwdpy11', ['COPYING', 'README.rst'])],
    long_description=long_desc,
    long_description_content_type='text/x-rst',
    ext_modules=ext_modules,
    install_requires=['pybind11>=2.4.3', 'numpy', 'tskit>=0.1.4', 'sparse'],
    cmdclass={'build_ext': CMakeBuild},
    packages=PKGS,
    package_data=generated_package_data,
    zip_safe=False,
)
